"""This file is part of asyncoro; see http://asyncoro.sourceforge.net for
details.

'discoro' implements generic coroutine scheduler that creates remote coroutines
at server processes. This scheduler is used by 'run' methods of
computations. This file implements special purpose schedulers using the same
'run' method.
"""

import inspect

import asyncoro.disasyncoro as asyncoro
import asyncoro.discoro as discoro
from asyncoro import Coro
from asyncoro.discoro import DiscoroStatus, DiscoroNodeInfo

__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__copyright__ = "Copyright (c) 2015 Giridhar Pemmasani"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"

__all__ = ['RemoteCoroScheduler']


class RemoteCoroScheduler(object):
    """Scheduler for submitting computation jobs.

    When coroutines are created with 'run' methods of Computation instances,
    they are created with a load-balancing algorithm on available servers with
    no limit on how many coroutines are run at a server. This works when
    coroutines are not CPU bound always. If, however, coroutines are
    computations (CPU bound always/mostly), then it may be more appropriate to
    schedule one coroutine at a server so creating a new coroutine waits until a
    server becomes available.

    RemoteCoroScheduler schedules at most one computation (coroutine) at a
    server process at any time (so a node may execute as many computation
    coroutines as there are server processes running on that node, but not
    more).

    See 'discomp*.py' files in 'examples' directory for some use cases.

    NB: When using this scheduler, 'run' method of computation shouldn't be used
    to create (remote) coroutines (unless those don't take up CPU), as this
    scheduler is not aware of those.
    """

    def __init__(self, computation, status=None, node_available=None,
                 proc_available=None, proc_close=None):
        """'computation' should be an instance of discoro.Computation

        'status' if not None should be a generator function that is called
        (as coroutine) with the status and info, as received by status_coro. If
        status is ServerInitialized and this function returns non-zero value,
        the server is ignored; i.e., jobs scheduled with 'schedule' or 'execute'
        will not use that server.

        'proc_available' if not None should be a generator function that is
        called (as coroutine) with the location of a server process when it
        becomes available (after all 'depends' of computation have been
        transferred). The coroutine runs at the client; it can create remote
        coroutine(s) at the server process, perhaps to setup, such as
        initializing global variables, transfer additional files etc. The
        coroutine should exit with 0 to indicate successful setup; any other
        value is interpretted as failure and not used by scheduler.

        'proc_close' if not None should be a generator function that is called
        (as coroutine) with the status and location of server process when
        server is about to be closed, or already closed. The coroutine runs at
        the client; it can create remote coroutine(s) at server process to
        cleanup, such as delete global variables, transfer files back to client
        etc. The coroutine is called with two parameters: 'status', which is
        either 'discoro.Scheduler.ServerInitialized' when server is about to be
        closed (i.e., server is still available, and remote coroutines can be
        executed), or 'discoro.Scheduler.ServerClosed' when server is already
        closed (e.g., due to zombie_period time elapsed without communication,
        or server was manually closed with command-line etc.), and 'location' of
        server process.
        """

        if status:
            if not inspect.isgeneratorfunction(status):
                asyncoro.logger.warning('Invalid status ignored')
                status = None
        if proc_available:
            if not inspect.isgeneratorfunction(proc_available):
                asyncoro.logger.warning('Invalid proc_available ignored')
                proc_available = None
        if proc_close:
            if not inspect.isgeneratorfunction(proc_close):
                asyncoro.logger.warning('Invalid proc_close ignored')
                proc_close = None
        if not node_available and computation._node_available:
            node_available = computation._node_available
        if node_available:
            if not inspect.isgeneratorfunction(node_available):
                asyncoro.logger.warning('Invalid node_available ignored')
                node_available = None

        self._status = status
        self._proc_available = proc_available
        self._proc_close = proc_close
        self._node_available = node_available
        self._close_servers = {}

        self.computation = computation
        self.computation_sign = None
        self.status_coro = Coro(self._status_proc)
        if isinstance(computation.status_coro, Coro):
            def chain_status_msgs(status_coro, client, coro=None):
                coro.set_daemon()
                while True:
                    msg = yield coro.receive()
                    client.send(msg)
                    status_coro.send(msg)
            computation.status_coro = Coro(chain_status_msgs, self.status_coro,
                                           computation.status_coro)
        else:
            computation.status_coro = self.status_coro
        self._rcoros = {}
        self._rcoros_done = asyncoro.Event()
        self._askew_results = {}
        self._servers = {}
        self._server_avail = asyncoro.Event()
        self._remote_scheduler = False
        self.asyncoro = asyncoro.AsynCoro()
        Coro(computation.schedule)

    def schedule(self, gen, *args, **kwargs):
        """Similar to 'run' method of computation, except as noted above: This
        method will block until a server process is available (i.e., not running
        another computation).

        Must be used with 'yield', similar to 'run' method of Computation
        instance.
        """

        while not self._servers:
            self._server_avail.clear()
            yield self._server_avail.wait()
        sloc, loc = self._servers.popitem()
        rcoro = yield self.computation.run_at(loc, gen, *args, **kwargs)
        if isinstance(rcoro, Coro):
            self._rcoros[rcoro] = (None, 1)
            if self._askew_results:
                msg = self._askew_results.pop(rcoro, None)
                if msg:
                    self.status_coro.send(msg)
        else:
            self._servers[sloc] = loc
            self._server_avail.set()
        raise StopIteration(rcoro)

    def execute(self, gen, *args, **kwargs):
        """Similar to 'run' method of computation, except as noted above: The
        caller (client coroutine) will block until a server process is available
        (i.e., not running another computation), where remote coroutine with
        given 'gen', 'args' and 'kwargs' runs and finishes. The return value is
        the result of computation.

        Must be used with 'yield', similar to 'run' method of Computation
        instance.
        """

        while not self._servers:
            self._server_avail.clear()
            yield self._server_avail.wait()
        sloc, loc = self._servers.popitem()
        rcoro = yield self.computation.run_at(loc, gen, *args, **kwargs)
        if isinstance(rcoro, Coro):
            client = asyncoro.AsynCoro.cur_coro()
            self._rcoros[rcoro] = (client, 1)
            if self._askew_results:
                msg = self._askew_results.pop(rcoro, None)
                if msg:
                    self.status_coro.send(msg)
            client._await_()
        else:
            self._servers[sloc] = loc
            self._server_avail.set()
            raise StopIteration(asyncoro.MonitorException(None, (type(rcoro), rcoro)))

    def execute_at(self, where, gen, *args, **kwargs):
        """Similar to 'run_at' method of computation, except the calling
        coroutine is blocked until the computation finishes and exit value of
        computation is returned. Unlike 'execute', the computation is executed
        right away, even if remote server process is executing another
        computation.

        Must be used with 'yield', similar to 'run_at' method of Computation
        instance.
        """

        rcoro = yield self.computation.run_at(where, gen, *args, **kwargs)
        if isinstance(rcoro, Coro):
            client = asyncoro.AsynCoro.cur_coro()
            self._rcoros[rcoro] = (client, 0)
            if self._askew_results:
                msg = self._askew_results.pop(rcoro, None)
                if msg:
                    self.status_coro.send(msg)
            client._await_()
        else:
            raise StopIteration(asyncoro.MonitorException(None, (type(rcoro), rcoro)))

    def map_results(self, gen, iter):
        """Execute generator 'gen' with arguments from given iterable. The
        return value is list of results that correspond to executing 'gen' with
        arguments in iterable in the same order.

        Must be used with 'yield', as for example,
        'results = yield scheduler.map_results(generator, list_of_tuples)'.
        """
        def exec_proc(gen, *args):
            yield self.execute(gen, *args)

        coros = []
        append_coro = coros.append
        for params in iter:
            if not isinstance(params, tuple):
                if hasattr(params, '__iter__'):
                    params = tuple(params)
                else:
                    params = (params,)
            append_coro(Coro(exec_proc, gen, *params))
        results = [None] * len(coros)
        for i, coro in enumerate(coros):
            result = yield coro.finish()
            results[i] = result
        raise StopIteration(results)

    def submit_at(self, where, gen, *args, **kwargs):
        """Similar to 'run_at' method of computation. If 'where' is None, the
        calling coroutine is blocked until any server is discovered and
        initialized (so computation's 'run_at' will not fail). Unlike
        'schedule', this method doesn't wait for server to be free (i.e., not
        running any other coroutines), nor unlike 'execute_at', the caller is
        not blocked until the coroutine finishes.

        Must be used with 'yield', similar to 'run_at' method of Computation
        instance. The value returned is result of 'run_at' method of computation
        (reference to remote coroutine in case of success, and error otherwise).
        """
        if not where:
            if not self._servers and not self._rcoros:
                yield self._server_avail.wait()
        rcoro = yield self.computation.run_at(where, gen, *args, **kwargs)
        if isinstance(rcoro, Coro):
            self._rcoros[rcoro] = (None, 0)
            if self._askew_results:
                msg = self._askew_results.pop(rcoro, None)
                if msg:
                    self.status_coro.send(msg)
        raise StopIteration(rcoro)

    def submit(self, gen, *args, **kwargs):
        """Submit coroutine at any server; see 'submit_at' above.
        """
        yield self.submit_at(None, gen, *args, **kwargs)

    def finish(self, close=False):
        """Wait until all scheduled coroutines finish. If 'close' is True, the
        computation is closed as well.

        Must be used with 'yield' as 'yield job_scheduler.finish()'.
        """

        self._rcoros_done.clear()
        if self._rcoros:
            yield self._rcoros_done.wait()
        if close:
            if self._proc_close:
                coros = [Coro(self._proc_close, discoro.Scheduler.ServerInitialized, location)
                         for location in self._close_servers]
                self._close_servers = {}
                for coro in coros:
                    yield coro.finish()
            else:
                self._close_servers = {}
        self._rcoros_done.clear()
        if self._rcoros:
            yield self._rcoros_done.wait()
        if close:
            yield self.computation.close()
            self._askew_results.clear()

    def _status_proc(self, coro=None):
        """Internal use only. Coroutine to process discoro scheduler messages.
        """

        coro.set_daemon()
        coro.scheduler().atexit(15, lambda: Coro(self.finish, True).value())
        while 1:
            msg = yield coro.receive()
            if isinstance(msg, asyncoro.MonitorException):
                if msg.args[1][0] == discoro.Scheduler.ServerClosed:
                    continue
                rcoro = msg.args[0]
                client, use_count = self._rcoros.pop(rcoro, ('missing', 0))
                if client is None:
                    pass
                elif isinstance(client, Coro):
                    client._proceed_(msg.args[1][1])
                elif client == 'missing':
                    # Due to 'yield' used to create rcoro, scheduler may not
                    # have updated self._rcoros before the coroutine's
                    # MonitorException is received, so put it in
                    # 'askew_results'. The scheduling coroutine will resend it
                    # when it receives rcoro
                    self._askew_results[rcoro] = msg
                    continue

                else:
                    asyncoro.logger.warning('RemoteCoroScheduler: invalid status message ignored')
                    continue
                if use_count:
                    self._servers[rcoro.location] = rcoro.location
                    self._server_avail.set()
                if not self._rcoros:
                    self._rcoros_done.set()

            elif isinstance(msg, DiscoroStatus):
                if msg.status == discoro.Scheduler.ServerInitialized:
                    if self._proc_available:
                        def setup_proc(self, msg, coro=None):
                            if self._remote_scheduler:
                                yield self.asyncoro.peer(msg.info)
                            if (yield Coro(self._proc_available, msg.info).finish()) == 0:
                                self._close_servers[msg.info] = msg.info
                                self._servers[msg.info] = msg.info
                                self._server_avail.set()
                        Coro(setup_proc, self, msg)
                    elif self._status:
                        def status_proc(self, msg, coro=None):
                            if (yield Coro(self._status, msg.status, msg.info).finish()) == 0:
                                self._servers[msg.info] = msg.info
                                self._server_avail.set()
                        Coro(status_proc, self, msg)
                    else:
                        self._servers[msg.info] = msg.info
                        self._server_avail.set()

                elif msg.status == discoro.Scheduler.ServerClosed:
                    self._servers.pop(msg.info, None)
                    if self._close_servers.pop(msg.info, None) and self._proc_close:
                        Coro(self._proc_close, msg.status, msg.info)
                    elif self._status:
                        Coro(self._status, msg.status, msg.info)
                elif msg.status == discoro.Scheduler.NodeDiscovered:
                    if self._node_available:
                        def setup_node(self, msg, coro=None):
                            if self._remote_scheduler:
                                yield self.asyncoro.peer(msg.info.location)
                            try:
                                params = yield asyncoro.Coro(self._node_available,
                                                             msg.info).finish()
                            except:
                                raise StopIteration

                            if not isinstance(params, tuple):
                                if hasattr(params, '__iter__'):
                                    params = tuple(params)
                                else:
                                    params = (params,)

                            msg = {'req': 'setup_node', 'addr': msg.info.location.addr,
                                   'params': params, 'auth': self.computation._auth,
                                   'client': coro}
                            self.computation.scheduler.send(msg)
                        Coro(setup_node, self, msg)
                elif msg.status == discoro.Scheduler.ComputationScheduled:
                    self.computation_sign = msg.info
                    if self.computation.scheduler.location != self.asyncoro.location:
                        self._remote_scheduler = True
                    if self._status:
                        Coro(self._status, msg.status, msg.info)
                elif (msg.status == discoro.Scheduler.ComputationClosed and
                      msg.info == self.computation_sign):
                    if self._status:
                        Coro(self._status, msg.status, msg.info)
                    raise StopIteration
                elif msg.status != discoro.Scheduler.CoroCreated:
                    if self._status:
                        Coro(self._status, msg.status, msg.info)

# This scheduler was called 'ProcScheduler' in earlier versions
ProcScheduler = RemoteCoroScheduler



# This example uses status messages and message passing to run 'setup' coroutine
# at remote process to prepare it for processing jobs.

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *
from asyncoro.discoro_schedulers import RemoteCoroScheduler

# Unlike in earlier versions of asyncoro, computations can now take time - even
# if computations don't "yield" to scheduler, asyncoro can still send/receive
# messages, respond to timer events in scheduler etc. In this case, computation
# is simulated with 'time.sleep' which blocks user asyncoro thread, but another
# (reactive) asycoro thread processes network traffic, run scheduler coroutines.

def compute_coro(coro=None):
    import time

    client = yield coro.receive() # first message is client coroutine

    result = 0
    while True:
        n = yield coro.receive()
        if n is None:  # end of requests
            client.send(result)
            break
        # long-running computation is simulated with 'time.sleep'; during this
        # time client may send messages to this coroutine, which will be
        # received and put in this coroutine's message queue (by _ReactAsynCoro_)
        time.sleep(n)
        result += n

# client (local) coroutine submits computations
def client_proc(computation, njobs, coro=None):
    # send 5 requests to remote process (compute_coro)
    def send_requests(rcoro, coro=None):
        # first send this local coroutine (to whom rcoro sends result)
        rcoro.send(coro)
        for i in range(5):
            # even if recipient doesn't use "yield" (such as executing long-run
            # computation, or thread-blocking function such as 'time.sleep' as
            # in this case), the message is accepted by another scheduler
            # (_ReactAsynCoro_) at the receiver and put in recipient's message
            # queue
            rcoro.send(random.uniform(10, 20))
            # assume delay in input availability
            yield coro.sleep(random.uniform(2, 5))
        # end of input is indicated with None
        rcoro.send(None)
        result = yield coro.receive() # get result
        print('    %s computed result: %.4f' % (rcoro.location, result))

    for i in range(njobs):
        rcoro = yield rcoro_scheduler.schedule(compute_coro)
        if isinstance(rcoro, asyncoro.Coro):
            print('  job %d processed by %s' % (i, rcoro.location))
            asyncoro.Coro(send_requests, rcoro)

    yield rcoro_scheduler.finish(close=True)


if __name__ == '__main__':
    import random, sys
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # if scheduler is not already running (on a node as a program), start
    # private scheduler:
    Scheduler()
    # package computation fragments
    computation = Computation([compute_coro])
    # use RemoteCoroScheduler to start coroutines at servers (should be done
    # before scheduling computation)
    rcoro_scheduler = RemoteCoroScheduler(computation)
    # run n jobs
    asyncoro.Coro(client_proc, computation, 10 if len(sys.argv) < 2 else int(sys.argv[1]))



#!/usr/bin/env python

# client and server coroutines communicating with message passing
# (asynchronous concurrent programming);
# see http://asyncoro.sourceforge.net/tutorial.html for details.

import sys, random
import asyncoro

def server_proc(coro=None):
    coro.set_daemon()
    while True:
        msg = yield coro.receive()
        print('processing %s' % (msg))

msg_id = 0

def client_proc(server, n, coro=None):
    global msg_id
    for x in range(3):
        yield coro.suspend(random.uniform(0.5, 3))
        msg_id += 1
        server.send('%d: %d / %d' % (msg_id, n, x))

# create server
server = asyncoro.Coro(server_proc)
# create 10 clients
for i in range(10):
    asyncoro.Coro(client_proc, server, i)



"""This file is part of asyncoro; see http://asyncoro.sourceforge.net
for details.

This module adds API for distributed programming to AsynCoro.
"""

import socket
import inspect
import traceback
import os
import stat
import hashlib
import collections
import copy
import tempfile
import threading
import errno
import atexit
try:
    import netifaces
except ImportError:
    netifaces = None

import asyncoro
from asyncoro import *

__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__copyright__ = "Copyright (c) 2012-2014 Giridhar Pemmasani"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"

__version__ = asyncoro.__version__
__all__ = asyncoro.__all__ + ['RCI']

# if connections to a peer are not successful consecutively
# MaxConnectionErrors times, peer is assumed dead and removed
MaxConnectionErrors = 10
MsgTimeout = asyncoro.MsgTimeout


class PeerStatus(object):
    """'peer_status' method of AsynCoro can be used to be notified of
    status of peers (other AsynCoro's to communicate for distributed
    programming). The status notifications are sent as messages to the
    regisered coroutine. Each message is an instance of this class.
    """

    Online = 1
    Offline = 0

    def __init__(self, location, name, status):
        self.location = location
        self.name = name
        self.status = status


class AsynCoro(asyncoro.AsynCoro, metaclass=Singleton):
    """This adds network services to asyncoro.AsynCoro so it can
    communicate with peers.

    If 'node' is not None, it must be either hostname or IP address
    where asyncoro runs network services. If 'udp_port' is not None,
    it is port number where asyncoro runs network services. If
    'udp_port' is 0, the default port number 51350 is used. If
    multiple instances of asyncoro are to be running on same host,
    they all can be started with the same 'udp_port', so that asyncoro
    instances automatically find each other.

    'name' is used in locating peers. They must be unique. If 'name'
    is not given, it is set to string 'node:tcp_port'.

    'ext_ip_addr' is the IP address of NAT firewall/gateway if
    asyncoro is behind that firewall/gateway.

    If 'discover_peers' is True (default), this node broadcasts message to
    detect other peers. If it is False, message is not broadcasted.

    'secret' is string that is used to hash which is used for
    authentication, so only peers that have same secret can
    communicate.

    'certfile' and 'keyfile' are path names for files containing SSL
    certificates; see Python 'ssl' module.

    'dest_path' is path to directory (folder) where transferred files
    are saved. If path doesn't exist, asyncoro creates directory with
    that path.

    'max_file_size' is maximum length of file in bytes allowed for
    transferred files. If it is 0 or None (default), there is no
    limit.
    """

    _instance = None
    _sys_asyncoro = None

    def __init__(self, *args, **kwargs):
        AsynCoro._instance = self
        atexit.register(self.finish)
        super(self.__class__, self).__init__()
        RCI._asyncoro = _SysAsynCoro_._asyncoro = self
        self._sys_asyncoro = _SysAsynCoro_(*args, **kwargs)
        self.__class__._sys_asyncoro = self._sys_asyncoro
        self._location = self._sys_asyncoro._location
        self._name = self._sys_asyncoro._name
        self._certfile = self._sys_asyncoro._certfile
        self._keyfile = self._sys_asyncoro._keyfile
        self.__dest_path_prefix = self.__dest_path = self._sys_asyncoro.dest_path

    @classmethod
    def instance(cls, *args, **kwargs):
        """Returns (singleton) instance of AsynCoro.
        """
        if not cls._instance:
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    @property
    def dest_path(self):
        return self.__dest_path

    @dest_path.setter
    def dest_path(self, path):
        if not self._sys_asyncoro:
            raise ValueError('AsynCoro not initialized!')
        path = os.path.normpath(path)
        if not path.startswith(self.__dest_path_prefix):
            path = os.path.join(self.__dest_path_prefix,
                                os.path.splitdrive(path)[1].lstrip(os.sep))
        try:
            os.makedirs(path)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise
        self._sys_asyncoro.dest_path = self.__dest_path = path

    def finish(self):
        if AsynCoro._instance:
            super(self.__class__, self).finish()
            AsynCoro._instance = None
            self._sys_asyncoro.finish()
            self._notifier.terminate()
            logger.shutdown()

    def locate(self, name, timeout=None):
        """Must be used with 'yield' as
        'loc = yield scheduler.locate("peer")'.

        Find and return location of peer with 'name'.
        """
        if not self._sys_asyncoro:
            raise StopIteration(None)
        _Peer._lock.acquire()
        for peer in _Peer.peers.values():
            if peer.name == name:
                loc = peer.location
                break
        else:
            loc = None
        _Peer._lock.release()
        if not loc:
            req = _NetRequest('locate_peer', kwargs={'name': name})
            req.event = Event()
            req_id = id(req)
            self._lock.acquire()
            self._pending_reqs[req_id] = req
            self._lock.release()
            _Peer.send_req_to(req, None)
            if (yield req.event.wait(timeout)) is False:
                req.reply = None
            loc = req.reply
            self._lock.acquire()
            self._pending_reqs.pop(req_id, None)
            self._lock.release()
        raise StopIteration(loc)

    def peer(self, loc, udp_port=0, stream_send=False, broadcast=False):
        """Must be used with 'yield', as
        'status = yield scheduler.peer("loc")'.

        Add asyncoro running at 'loc' as peer to communicate. Peers on
        a local network can find each other automatically, but if they
        are on different networks, 'peer' can be used so they find
        each other. 'loc' can be either an instance of Location or
        host name or IP address. If 'loc' is Location instance and
        'port' is 0, or 'loc' is host name or IP address, then all
        asyncoros running at the host will have streaming mode set as
        per 'stream_send'.

        If 'stream_send' is True, this asyncoro uses same connection
        again and again to send messages (i.e., as a stream) to peer
        'host' (instead of one message per connection).

        If 'broadcast' is True, the client information is broadcast on
        the network of peer. This can be used if client is on remote
        network and needs to communicate with all asyncoro's available
        on the network of peer (at 'loc').
        """

        if not self._sys_asyncoro:
            raise StopIteration(-1)

        def _peer(coro=None):
            SysCoro(self._sys_asyncoro.peer, coro, loc, udp_port, stream_send, broadcast)
            yield coro.recv()

        yield Coro(_peer).finish()

    def peer_status(self, coro):
        """This method can be used to be notified of status of peers
        (other AsynCoro's to communicate for distributed
        programming). The status notifications are sent as messages to
        the regisered coroutine. Each message is an instance of
        PeerStatus.
        """
        _Peer.peer_status(coro)

    def peers(self):
        """Returns list of current peers (as Location instances).
        """
        return _Peer.get_peers()

    def close_peer(self, location, timeout=MsgTimeout):
        """Must be used with 'yield', as
        'yield scheduler.close_peer("loc")'.

        Close peer at 'location'.
        """
        if isinstance(location, Location):
            _Peer._lock.acquire()
            peer = _Peer.peers.get((location.addr, location.port), None)
            _Peer._lock.release()
            if peer:

                def sys_proc(peer, timeout, done, coro=None):
                    yield _Peer.close_peer(peer, timeout=timeout, coro=coro)
                    done.set()

                event = Event()
                SysCoro(sys_proc, peer, timeout, event)
                yield event.wait()

    def ignore_peers(self, ignore):
        """Don't respond to 'ping' from peers if 'ignore=True'. This can be used
        during shutdown, or to limit peers to communicate.
        """
        if self._sys_asyncoro:
            self._sys_asyncoro.ignore_peers(ignore)

    def discover_peers(self, port=None):
        """This method can be invoked (periodically?) to broadcast message to
        discover peers, if there is a chance initial broadcast message may be
        lost (as these messages are sent over UDP).
        """
        if not self._sys_asyncoro:
            return
        SysCoro(self._sys_asyncoro.discover_peers, port=port)

    def send_file(self, location, file, dir=None, overwrite=False, timeout=MsgTimeout):
        """Must be used with 'yield' as
        'val = yield scheduler.send_file(location, "file1")'.

        Transfer 'file' to peer at 'location'. If 'dir' is not None, it must be
        a relative path (not absolute path), in which case, file will be saved
        at peer's dest_path + dir. Returns -1 in case of error, 0 if the file is
        transferred, 1 if the same file is already at the destination with same
        size, timestamp and permissions (so file is not transferred) and os.stat
        structure if a file with same name is at the destination with different
        size/timestamp/permissions, but 'overwrite' is False. 'timeout' is max
        seconds to transfer 1MB of data. If return value is 0, the sender may
        want to delete file with 'del_file' later.
        """
        try:
            stat_buf = os.stat(file)
        except:
            logger.warning('send_file: File "%s" is not valid', file)
            raise StopIteration(-1)
        if not ((stat.S_IMODE(stat_buf.st_mode) & stat.S_IREAD) and stat.S_ISREG(stat_buf.st_mode)):
            logger.warning('send_file: File "%s" is not valid', file)
            raise StopIteration(-1)
        if dir:
            if not isinstance(dir, str):
                logger.warning('send_file: path for dir "%s" is not allowed', dir)
                raise StopIteration(-1)
            dir = dir.strip()
            # reject absolute path for dir
            if os.path.join(os.sep, dir) == dir:
                logger.warning('send_file: Absolute path for dir "%s" is not allowed', dir)
                raise StopIteration(-1)
        else:
            dir = None
        peer = _Peer.get_peer(location)
        if peer is None:
            logger.debug('%s is not a valid peer', location)
            raise StopIteration(-1)
        kwargs = {'file': os.path.basename(file), 'stat_buf': stat_buf,
                  'overwrite': overwrite is True, 'dir': dir, 'sep': os.sep}
        req = _NetRequest('send_file', kwargs=kwargs, dst=location, timeout=timeout)
        sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
                           keyfile=self._keyfile, certfile=self._certfile)
        if timeout:
            sock.settimeout(timeout)
        fd = open(file, 'rb')
        try:
            yield sock.connect((location.addr, location.port))
            req.auth = peer.auth
            yield sock.send_msg(serialize(req))
            recvd = yield sock.recv_msg()
            recvd = deserialize(recvd)
            sent = 0
            while sent == recvd:
                data = fd.read(1024000)
                if not data:
                    break
                yield sock.sendall(data)
                sent += len(data)
                recvd = yield sock.recv_msg()
                recvd = deserialize(recvd)
            if recvd == stat_buf.st_size:
                reply = 0
            else:
                reply = -1
        except socket.error as exc:
            reply = -1
            logger.debug('could not send "%s" to %s', req.name, location)
            if len(exc.args) == 1 and exc.args[0] == 'hangup':
                logger.warning('peer "%s" not reachable', location)
                # TODO: remove peer?
        except:
            logger.warning('send_file: Could not send "%s" to %s', file, location)
            reply = -1
        finally:
            sock.close()
            fd.close()
        raise StopIteration(reply)

    def del_file(self, location, file, dir=None, timeout=None):
        """Must be used with 'yield' as
        'loc = yield scheduler.del_file(location, "file1")'.

        Delete 'file' from peer at 'location'. 'dir' must be
        same as that used for 'send_file'.
        """
        if isinstance(dir, str) and dir:
            dir = dir.strip()
            # reject absolute path for dir
            if os.path.join(os.sep, dir) == dir:
                raise StopIteration(-1)
        kwargs = {'file': os.path.basename(file), 'dir': dir}
        req = _NetRequest('del_file', kwargs=kwargs, dst=location, timeout=timeout)
        reply = yield _Peer._sync_reply(req)
        if reply is None:
            reply = -1
        raise StopIteration(reply)

    def _sys_call_(self, method, *args, **kwargs):
        swing = {'result': None, 'event': Event()}
        SysCoro(self._sys_asyncoro._swing_call_, swing, method, *args, **kwargs)
        yield swing['event'].wait()
        yield swing['result']


class RCI(object):
    """Remote Coro (Callable) Interface.

    Methods registered with RCI can be executed as coroutines on
    request (by remotely running coroutines).
    """

    __slots__ = ('_name', '_location', '_method')

    _asyncoro = None

    def __init__(self, method, name=None):
        """'method' must be generator method; this is used to create
        coroutines. If 'name' is not given, method's function name is
        used for registering.
        """
        if not inspect.isgeneratorfunction(method):
            raise RuntimeError('method must be generator function')
        self._method = method
        if name:
            self._name = name
        else:
            self._name = method.__name__
        if not RCI._asyncoro:
            RCI._asyncoro = AsynCoro.instance()
        self._location = RCI._asyncoro._location

    @property
    def location(self):
        """Get Location instance where this RCI is running.
        """
        return copy.copy(self._location)

    @property
    def name(self):
        """Get name of RCI.
        """
        return self._name

    @staticmethod
    def locate(name, location=None, timeout=None):
        """Must be used with 'yield' as
        'rci = yield RCI.locate("name")'.

        Returns RCI instance to registered RCI at a remote peer so
        its method can be used to execute coroutines at that peer.

        If 'location' is given, RCI is looked up at that specific
        peer; otherwise, all known peers are queried for given name.
        """
        if not RCI._asyncoro:
            RCI._asyncoro = AsynCoro.instance()
        req = _NetRequest('locate_rci', kwargs={'name': name}, dst=location, timeout=timeout)
        req.event = Event()
        req_id = id(req)
        RCI._asyncoro._lock.acquire()
        RCI._asyncoro._pending_reqs[req_id] = req
        RCI._asyncoro._lock.release()
        _Peer.send_req_to(req, location)
        if (yield req.event.wait(timeout)) is False:
            req.reply = None
        rci = req.reply
        RCI._asyncoro._lock.acquire()
        RCI._asyncoro._pending_reqs.pop(req_id, None)
        RCI._asyncoro._lock.release()
        raise StopIteration(rci)

    def register(self):
        """RCI must be registered so it can be located.
        """
        if self._location != RCI._asyncoro._location:
            return -1
        if not inspect.isgeneratorfunction(self._method):
            return -1
        RCI._asyncoro._lock.acquire()
        if RCI._asyncoro._rcis.get(self._name, None) is None:
            RCI._asyncoro._rcis[self._name] = self
            RCI._asyncoro._lock.release()
            return 0
        else:
            RCI._asyncoro._lock.release()
            return -1

    def unregister(self):
        """Unregister registered RCI; see 'register' above.
        """
        if self._location != RCI._asyncoro._location:
            return -1
        RCI._asyncoro._lock.acquire()
        if RCI._asyncoro._rcis.pop(self._name, None) is None:
            RCI._asyncoro._lock.release()
            return -1
        else:
            RCI._asyncoro._lock.release()
            return 0

    def __call__(self, *args, **kwargs):
        """Must be used with 'yeild' as 'rcoro = yield rci(*args, **kwargs)'.

        Run RCI (method at remote location) with args and kwargs. Both
        args and kwargs must be serializable. Returns (remote) Coro
        instance.
        """
        req = _NetRequest('run_rci', kwargs={'name': self._name, 'args': args, 'kwargs': kwargs},
                          dst=self._location, timeout=MsgTimeout)
        reply = yield _Peer._sync_reply(req)
        if isinstance(reply, Coro):
            raise StopIteration(reply)
        elif reply is None:
            raise StopIteration(None)
        else:
            raise Exception(reply)

    def __getstate__(self):
        state = {'_name': self._name, '_location': self._location}
        return state

    def __setstate__(self, state):
        self._name = state['_name']
        self._location = state['_location']

    def __eq__(self, other):
        return (isinstance(other, RCI) and
                self._name == other._name and self._location == other._location)

    def __ne__(self, other):
        return ((not isinstance(other, RCI)) or
                self._name != other._name or self._location != other._location)

    def __repr__(self):
        s = '%s' % (self._name)
        if self._location:
            s = '%s@%s' % (s, self._location)
        return s


class SysCoro(asyncoro.Coro):
    """Coroutine meant for reactive components, always ready to respond to
    events; i.e., takes very little CPU time to process events. This is meant
    for internal use to handle network traffic, timer events etc. in asyncoro
    modules. These coroutines run in seperate AsynCoro thread (_SysAsynCoro_),
    so if user coroutines (Coro instances) take too much CPU time, SysCoro can
    still respond to such events quickly.

    This class is not meant for users, as using this improperly may break
    (parts of) asyncoro.
    """

    _asyncoro = None

    def __init__(self, *args, **kwargs):
        if not SysCoro._asyncoro:
            AsynCoro.instance()
        self._scheduler = SysCoro._asyncoro
        super(SysCoro, self).__init__(*args, **kwargs)


class _NetRequest(object):
    """Internal use only.
    """

    __slots__ = ('name', 'kwargs', 'dst', 'auth', 'event', 'reply', 'timeout')

    def __init__(self, name, kwargs={}, dst=None, auth=None, timeout=None):
        self.name = name
        self.kwargs = kwargs
        self.dst = dst
        self.auth = auth
        self.event = None
        self.reply = None
        self.timeout = timeout

    def __getstate__(self):
        state = {'name': self.name, 'kwargs': self.kwargs, 'dst': self.dst,
                 'auth': self.auth, 'reply': self.reply, 'timeout': self.timeout}
        return state

    def __setstate__(self, state):
        for k, v in state.items():
            setattr(self, k, v)


class _Peer(object):
    """Internal use only.
    """

    __slots__ = ('name', 'location', 'auth', 'keyfile', 'certfile', 'stream', 'conn',
                 'reqs', 'waiting', 'req_coro')

    peers = {}
    status_coro = None
    _asyncoro = None
    _lock = threading.Lock()

    def __init__(self, name, location, auth, keyfile, certfile):
        self.name = name
        self.location = location
        self.auth = auth
        self.keyfile = keyfile
        self.certfile = certfile
        self.stream = False
        self.conn = None
        self.reqs = collections.deque()
        self.waiting = False
        _Peer._lock.acquire()
        _Peer.peers[(location.addr, location.port)] = self
        _Peer._lock.release()
        self.req_coro = SysCoro(self.req_proc)
        logger.debug('%s: found peer %s', _Peer._asyncoro._location, self.location)
        if _Peer.status_coro:
            _Peer.status_coro.send(PeerStatus(location, name, PeerStatus.Online))

        _SysAsynCoro_._asyncoro._lock.acquire()
        if ((location.addr, location.port) in _SysAsynCoro_._asyncoro._stream_peers or
            (location.addr, 0) in _SysAsynCoro_._asyncoro._stream_peers):
            self.stream = True

        # send pending (async) requests
        for pending_req in _SysAsynCoro_._asyncoro._pending_reqs.values():
            if (pending_req.name == 'locate_peer' and pending_req.kwargs['name'] == self.name):
                pending_req.reply = location
                pending_req.event.set()

            if pending_req.dst:
                if pending_req.dst == location:
                    _Peer.send_req(pending_req)
            else:
                _Peer.send_req_to(pending_req, location)
        _SysAsynCoro_._asyncoro._lock.release()

    @staticmethod
    def get_peers():
        _Peer._lock.acquire()
        peers = [copy.copy(peer.location) for peer in _Peer.peers.values()]
        _Peer._lock.release()
        return peers

    @staticmethod
    def get_peer(location):
        _Peer._lock.acquire()
        peer = _Peer.peers.get((location.addr, location.port), None)
        _Peer._lock.release()
        return peer

    @staticmethod
    def send_req(req):
        _Peer._lock.acquire()
        peer = _Peer.peers.get((req.dst.addr, req.dst.port), None)
        if not peer:
            logger.debug('%s: invalid peer: %s', _Peer._asyncoro.location, req.dst)
            _Peer._lock.release()
            return -1
        peer.reqs.append(req)
        if peer.waiting:
            peer.waiting = False
            peer.req_coro.send(1)
        _Peer._lock.release()
        return 0

    @staticmethod
    def send_req_to(req, dst):
        if dst:
            _Peer._lock.acquire()
            peer = _Peer.peers.get((dst.addr, dst.port), None)
            if not peer:
                logger.debug('%s: invalid peer to: %s', _Peer._asyncoro.location, dst)
                _Peer._lock.release()
                return -1
            peer.reqs.append(req)
            if peer.waiting:
                peer.waiting = False
                peer.req_coro.send(1)
            _Peer._lock.release()
        else:
            _Peer._lock.acquire()
            for peer in _Peer.peers.values():
                peer.reqs.append(req)
                if peer.waiting:
                    peer.waiting = False
                    peer.req_coro.send(1)
            _Peer._lock.release()
        return 0

    @staticmethod
    def _sync_reply(req, alarm_value=None):
        req.event = Event()
        if _Peer.send_req(req) != 0:
            raise StopIteration(-1)
        if (yield req.event.wait(req.timeout)) is False:
            raise StopIteration(alarm_value)
        raise StopIteration(req.reply)

    @staticmethod
    def close_peer(peer, timeout, coro=None):
        req = _NetRequest('close_peer', kwargs={'location': _Peer._asyncoro._location},
                          dst=peer.location, timeout=timeout)
        yield _Peer._sync_reply(req)
        if peer.req_coro:
            yield peer.req_coro.terminate()
            while peer.req_coro:
                yield coro.sleep(0.1)

    @staticmethod
    def shutdown(timeout=MsgTimeout):
        _Peer._lock.acquire()
        for peer in _Peer.peers.values():
            SysCoro(_Peer.close_peer, peer, timeout)
        _Peer._lock.release()

    def req_proc(self, coro=None):
        coro.set_daemon()
        conn_errors = 0
        req = None
        while 1:
            _Peer._lock.acquire()
            if self.reqs:
                _Peer._lock.release()
            else:
                self.waiting = True
                _Peer._lock.release()
                if not self.stream and self.conn:
                    self.conn.shutdown(socket.SHUT_WR)
                    self.conn.close()
                    self.conn = None
                try:
                    yield coro.receive()
                except GeneratorExit:
                    break
            req = self.reqs.popleft()
            if not self.conn:
                self.conn = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
                                        keyfile=self.keyfile, certfile=self.certfile)
                if req.timeout:
                    self.conn.settimeout(req.timeout)
                try:
                    yield self.conn.connect((self.location.addr, self.location.port))
                except GeneratorExit:
                    if self.conn:
                        try:
                            self.conn.shutdown(socket.SHUT_WR)
                            self.conn.close()
                        except:
                            pass
                        self.conn = None
                    break
                except:
                    if self.conn:
                        # self.conn.shutdown(socket.SHUT_WR)
                        self.conn.close()
                        self.conn = None
                    req.reply = None
                    if req.event:
                        req.event.set()
                    conn_errors += 1
                    if conn_errors >= MaxConnectionErrors:
                        logger.warning('too many connection errors to %s; removing it',
                                       self.location)
                        break
                    continue
                else:
                    if conn_errors:
                        conn_errors = 0
            else:
                self.conn.settimeout(req.timeout)

            req.auth = self.auth
            try:
                yield self.conn.send_msg(serialize(req))
                reply = yield self.conn.recv_msg()
                reply = deserialize(reply)
                if req.event:
                    if reply is not None or req.dst == self.location:
                        req.reply = reply
                        req.event.set()
                else:
                    req.reply = reply
            except socket.error as exc:
                logger.debug('%s: Could not send "%s" to %s', _Peer._asyncoro._location, req.name,
                             self.location)
                # logger.debug(traceback.format_exc())
                if len(exc.args) == 1 and exc.args[0] == 'hangup':
                    logger.warning('peer "%s" not reachable', self.location)
                    # TODO: remove peer?
                try:
                    self.conn.shutdown(socket.SHUT_WR)
                    self.conn.close()
                except:
                    pass
                self.conn = None
                req.reply = None
                if req.event:
                    req.event.set()
            except socket.timeout:
                # logger.debug(traceback.format_exc())
                try:
                    self.conn.shutdown(socket.SHUT_WR)
                    self.conn.close()
                except:
                    pass
                self.conn = None
                req.reply = None
                if req.event:
                    req.event.set()
            except GeneratorExit:
                if self.conn:
                    try:
                        self.conn.shutdown(socket.SHUT_WR)
                        self.conn.close()
                    except:
                        pass
                    self.conn = None
                break
            except:
                # logger.debug(traceback.format_exc())
                if self.conn:
                    try:
                        self.conn.shutdown(socket.SHUT_WR)
                        self.conn.close()
                    except:
                        pass
                    self.conn = None
                req.reply = None
                if req.event:
                    req.event.set()

        if req and isinstance(req.event, Event):
            req.reply = None
            req.event.set()
        for req in self.reqs:
            if isinstance(req.event, Event):
                req.reply = None
                req.event.set()

        self.reqs.clear()
        self.req_coro = None
        if self.conn:
            # self.conn.shutdown(socket.SHUT_WR)
            self.conn.close()
            self.conn = None
        _Peer.remove(self.location)
        raise StopIteration(None)

    @staticmethod
    def remove(location):
        _Peer._lock.acquire()
        peer = _Peer.peers.pop((location.addr, location.port), None)
        _Peer._lock.release()
        if peer:
            logger.debug('%s: peer %s terminated', _Peer._asyncoro.location, peer.location)
            peer.stream = False
            if peer.req_coro:
                peer.req_coro.terminate()
            if _Peer.status_coro:
                _Peer.status_coro.send(PeerStatus(peer.location, peer.name, PeerStatus.Offline))

    @staticmethod
    def peer_status(coro):
        _Peer._lock.acquire()
        if isinstance(coro, Coro):
            # if there is another status_coro, add or replace?
            for peer in _Peer.peers.values():
                try:
                    coro.send(PeerStatus(peer.location, peer.name, PeerStatus.Online))
                except:
                    logger.debug(traceback.format_exc())
                    break
            else:
                _Peer.status_coro = coro
        elif coro is None:
            _Peer.status_coro = None
        else:
            logger.warning('invalid peer status coroutine ignored')
        _Peer._lock.release()


class _SysAsynCoro_(asyncoro.AsynCoro, metaclass=Singleton):
    """Internal use only.
    """

    _instance = None
    _asyncoro = None

    def __init__(self, udp_port=0, tcp_port=0, node=None, ext_ip_addr=None,
                 name=None, discover_peers=True,
                 secret='', certfile=None, keyfile=None, notifier=None,
                 dest_path=None, max_file_size=None):
        super(self.__class__, self).__init__()
        SysCoro._asyncoro = _Peer._asyncoro = self
        if node:
            node = socket.gethostbyname(node)
        else:
            if netifaces:
                for iface in netifaces.interfaces():
                    for link in netifaces.ifaddresses(iface).get(netifaces.AF_INET, []):
                        if link.get('broadcast', None) and link.get('netmask', None):
                            node = socket.gethostbyname(link.get('addr', ''))
                            break
                    else:
                        continue
                    break
            if not node:
                node = socket.gethostbyname(socket.gethostname())
        if not udp_port:
            udp_port = 51350
        if not dest_path:
            dest_path = os.path.join(os.sep, tempfile.gettempdir(), 'asyncoro')
        self.__dest_path = os.path.abspath(os.path.normpath(dest_path))
        self.__dest_path_prefix = dest_path
        # TODO: avoid race condition (use locking to check/create atomically?)
        if not os.path.isdir(self.__dest_path):
            try:
                os.makedirs(self.__dest_path)
            except:
                # likely another asyncoro created this directory
                if not os.path.isdir(self.__dest_path):
                    logger.warning('failed to create "%s"', self.__dest_path)
                    logger.debug(traceback.format_exc())
        self.max_file_size = max_file_size
        self._certfile = certfile
        self._keyfile = keyfile
        self._udp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
        if hasattr(socket, 'SO_REUSEADDR'):
            self._udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(socket, 'SO_REUSEPORT'):
            self._udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self._udp_sock.bind(('', udp_port))
        self._tcp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
                                     keyfile=self._keyfile, certfile=self._certfile)
        if tcp_port:
            if hasattr(socket, 'SO_REUSEADDR'):
                self._tcp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if hasattr(socket, 'SO_REUSEPORT'):
                self._tcp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self._tcp_sock.bind((node, tcp_port))
        self._location = Location(*self._tcp_sock.getsockname())
        if not self._location.port:
            raise Exception('could not start network server at %s' % (self._location))
        if name:
            self._name = name
        else:
            self._name = str(self._location)
        if ext_ip_addr:
            try:
                ext_ip_addr = socket.gethostbyname(ext_ip_addr)
            except:
                logger.warning('invalid ext_ip_addr ignored')
            else:
                self._location.addr = ext_ip_addr

        self._secret = secret
        if secret is None:
            self._signature = None
            self._auth_code = None
        else:
            self._signature = ''.join(hex(_)[2:] for _ in os.urandom(20))
            self._auth_code = hashlib.sha1((self._signature + secret).encode()).hexdigest()
        self._tcp_sock.listen(32)
        logger.info('network server %s@ %s, udp_port=%s', '"%s" ' % name if name else '',
                    self._location, self._udp_sock.getsockname()[1])
        self._broadcast = '<broadcast>'
        if netifaces:
            for iface in netifaces.interfaces():
                for link in netifaces.ifaddresses(iface).get(netifaces.AF_INET, []):
                    if link['addr'] == self._location.addr:
                        self._broadcast = link.get('broadcast', '<broadcast>')
                        break
                else:
                    continue
                break
        self._ignore_peers = False
        self._tcp_coro = SysCoro(self._tcp_proc)
        self._udp_coro = SysCoro(self._udp_proc, discover_peers)

    @staticmethod
    def instance():
        return _SysAsynCoro_._instance

    @property
    def dest_path(self):
        return self.__dest_path

    @dest_path.setter
    def dest_path(self, path):
        self.__dest_path = path

    def finish(self):
        super(self.__class__, self).finish()
        if self._udp_sock:
            self._udp_sock.close()
            self._udp_sock = None
        if self._tcp_sock:
            self._tcp_sock.close()
            self._tcp_sock = None

    def peer(self, client, loc, udp_port=0, stream_send=False, broadcast=False, coro=None):
        """
        _Must_ be called with SysCoro
        """
        def _peer(loc, udp_port, stream_send, broadcast):
            if not isinstance(loc, Location):
                try:
                    loc = socket.gethostbyname(loc)
                except:
                    logger.warning('invalid node: "%s"', str(loc))
                    raise StopIteration(-1)
                loc = Location(loc, 0)

            _SysAsynCoro_._asyncoro._lock.acquire()
            if stream_send:
                self._stream_peers[(loc.addr, loc.port)] = True
            else:
                self._stream_peers.pop((loc.addr, loc.port), None)

            if loc.port:
                _Peer._lock.acquire()
                peer = _Peer.peers.get((loc.addr, loc.port), None)
                _Peer._lock.release()
                if peer:
                    peer.stream = stream_send
                    if not broadcast:
                        _SysAsynCoro_._asyncoro._lock.release()
                        raise StopIteration(0)
            else:
                _Peer._lock.acquire()
                for (addr, port), peer in _Peer.peers.items():
                    if addr == loc.addr:
                        peer.stream = stream_send
                        if not stream_send:
                            self._stream_peers.pop((addr, port), None)
                _Peer._lock.release()
            _SysAsynCoro_._asyncoro._lock.release()

            if loc.port:
                req = _NetRequest('signature', kwargs={'version': __version__}, dst=loc)
                sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
                sock.settimeout(2)
                try:
                    yield sock.connect((loc.addr, loc.port))
                    yield sock.send_msg(serialize(req))
                    sign = yield sock.recv_msg()
                    ret = yield self._acquaint_(loc, sign.decode(), coro=coro)
                except:
                    ret = -1
                finally:
                    sock.close()

                if broadcast and ret == 0:
                    kwargs = {'location': self._location, 'signature': self._signature,
                              'name': self._name, 'version': __version__}
                    _Peer.send_req_to(_NetRequest('relay_ping', kwargs=kwargs, dst=loc), loc)
                raise StopIteration(ret)
            else:
                if not udp_port:
                    udp_port = 51350
                ping_msg = {'location': self._location, 'signature': self._signature,
                            'name': self._name, 'version': __version__, 'broadcast': broadcast}
                ping_msg = 'ping:'.encode() + serialize(ping_msg)
                sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
                sock.settimeout(2)
                try:
                    yield sock.sendto(ping_msg, (loc.addr, udp_port))
                except:
                    pass
                sock.close()
            raise StopIteration(0)

        ret = yield _peer(loc, udp_port, stream_send, broadcast)
        client.send(ret)

    def _acquaint_(self, peer_location, peer_signature, coro=None):
        sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
                           keyfile=self._keyfile, certfile=self._certfile)
        sock.settimeout(MsgTimeout)
        req = _NetRequest('peer', kwargs={'signature': self._signature, 'name': self._name,
                                          'from': self._location, 'version': __version__},
                          dst=peer_location)
        req.auth = hashlib.sha1((peer_signature + self._secret).encode()).hexdigest()
        try:
            yield sock.connect((peer_location.addr, peer_location.port))
            yield sock.send_msg(serialize(req))
            peer_info = yield sock.recv_msg()
            peer_info = deserialize(peer_info)
            assert peer_info['version'] == __version__
            _Peer(peer_info['name'], peer_location, req.auth, self._keyfile, self._certfile)
            reply = 0
        except:
            logger.debug(traceback.format_exc())
            reply = -1
        sock.close()
        raise StopIteration(0)

    def discover_peers(self, port=None, coro=None):
        ping_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
        ping_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        ping_sock.settimeout(2)
        ping_sock.bind((self._tcp_sock.getsockname()[0], 0))
        ping_msg = {'location': self._location, 'signature': self._signature,
                    'name': self._name, 'version': __version__}
        ping_msg = 'ping:'.encode() + serialize(ping_msg)
        if not port:
            port = self._udp_sock.getsockname()[1]
        try:
            yield ping_sock.sendto(ping_msg, (self._broadcast, port))
        except:
            pass
        ping_sock.close()

    def ignore_peers(self, ignore):
        if ignore:
            self._ignore_peers = True
        else:
            self._ignore_peers = False

    def _udp_proc(self, discover_peers, coro=None):
        coro.set_daemon()
        if discover_peers:
            SysCoro(self.discover_peers)

        while 1:
            msg, addr = yield self._udp_sock.recvfrom(1024)
            if not msg.startswith(b'ping:'):
                logger.warning('ignoring UDP message from %s:%s', addr[0], addr[1])
                continue
            try:
                ping_info = deserialize(msg[len(b'ping:'):])
            except:
                continue
            peer_location = ping_info.get('location', None)
            if not isinstance(peer_location, Location) or peer_location == self._location:
                continue
            if ping_info['version'] != __version__:
                logger.warning('Peer %s version %s is not %s',
                               peer_location, ping_info['version'], __version__)
                continue
            if self._ignore_peers:
                continue
            if self._secret is None:
                auth_code = None
            else:
                auth_code = hashlib.sha1((ping_info.get('signature', '') +
                                          self._secret).encode()).hexdigest()
            _Peer._lock.acquire()
            peer = _Peer.peers.get((peer_location.addr, peer_location.port), None)
            _Peer._lock.release()
            if peer and peer.auth != auth_code:
                _Peer.remove(peer_location)
                peer = None

            if not peer:
                SysCoro(self._acquaint_, peer_location, ping_info['signature'])

    def _tcp_proc(self, coro=None):
        coro.set_daemon()
        while 1:
            conn, addr = yield self._tcp_sock.accept()
            SysCoro(self._tcp_task, conn, addr)

    def _tcp_task(self, conn, addr, coro=None):
        while 1:
            try:
                msg = yield conn.recv_msg()
            except:
                break
            if not msg:
                break
            try:
                req = deserialize(msg)
            except:
                logger.debug('%s ignoring invalid message', self._location)
                break
            if req.auth != self._auth_code:
                if req.name == 'signature':
                    yield conn.send_msg(self._signature.encode())
                else:
                    logger.warning('invalid request "%s" ignored', req.name)
                break

            # if req.dst and req.dst != self._location:
            #     logger.debug('invalid request "%s" to %s (%s)', req.name, req.dst, self._location)
            #     break

            if req.name == 'send':
                # synchronous message
                reply = -1
                if req.dst != self._location:
                    logger.warning('ignoring invalid "send" (%s != %s)', req.dst, self._location)
                else:
                    coro = req.kwargs.get('coro', None)
                    if coro:
                        name = req.kwargs.get('name', ' ')
                        if name[0] == '~':
                            Coro._asyncoro._lock.acquire()
                            coro = Coro._asyncoro._coros.get(int(coro), None)
                            Coro._asyncoro._lock.release()
                            if coro and coro._name == name:
                                reply = coro.send(req.kwargs['message'])
                            else:
                                logger.warning('ignoring invalid recipient to "send"')
                        elif name[0] == '!':
                            coro = self._coros.get(int(coro))
                            if coro and coro._name == name:
                                reply = coro.send(req.kwargs['message'])
                            else:
                                logger.warning('ignoring invalid recipient to "send"')
                        else:
                            logger.warning('invalid "send" message ignored')
                    else:
                        channel = req.kwargs.get('channel', None)
                        if channel[0] == '~':
                            Channel._asyncoro._lock.acquire()
                            channel = Channel._asyncoro._channels.get(channel)
                            Channel._asyncoro._lock.release()
                            if channel:
                                reply = channel.send(req.kwargs['message'])
                            else:
                                logger.warning('ignoring invalid recipient to "send"')
                        elif channel[0] == '!':
                            channel = self._channels.get(channel)
                            if isinstance(channel, Channel):
                                reply = channel.send(req.kwargs['message'])
                            else:
                                logger.warning('invalid "send" message ignored')
                        else:
                            logger.warning('ignoring invalid recipient to "send"')
                yield conn.send_msg(serialize(reply))

            elif req.name == 'deliver':
                # synchronous message
                reply = -1
                if req.dst != self._location:
                    logger.warning('ignoring invalid "deliver" (%s != %s)', req.dst, self._location)
                else:
                    coro = req.kwargs.get('coro', None)
                    if coro:
                        name = req.kwargs.get('name', ' ')
                        if name[0] == '~':
                            Coro._asyncoro._lock.acquire()
                            coro = Coro._asyncoro._coros.get(int(coro))
                            Coro._asyncoro._lock.release()
                            if coro and coro.send(req.kwargs['message']) == 0:
                                reply = 1
                        elif name[0] == '!':
                            coro = self._coros.get(int(coro))
                            if coro and coro.send(req.kwargs['message']) == 0:
                                reply = 1
                            else:
                                logger.warning('invalid "deliver" message ignored')
                    else:
                        channel = req.kwargs.get('channel')
                        if channel:
                            if channel[0] == '~':
                                Channel._asyncoro._lock.acquire()
                                channel = Channel._asyncoro._channels.get(channel)
                                Channel._asyncoro._lock.release()
                                if channel:
                                    reply = yield channel.deliver(
                                        req.kwargs['message'], timeout=req.timeout,
                                        n=req.kwargs['n'])
                            elif channel[0] == '!':
                                channel = self._channels.get(channel)
                                if isinstance(channel, Channel):
                                    reply = yield channel.deliver(
                                        req.kwargs['message'], timeout=req.timeout,
                                        n=req.kwargs['n'])
                            else:
                                logger.warning('invalid "deliver" message ignored')
                        else:
                            logger.warning('invalid "deliver" message ignored')
                yield conn.send_msg(serialize(reply))

            elif req.name == 'run_rci':
                # synchronous message
                if req.dst != self._location:
                    reply = Exception('invalid RCI invocation')
                else:
                    RCI._asyncoro._lock.acquire()
                    rci = RCI._asyncoro._rcis.get(req.kwargs['name'], None)
                    RCI._asyncoro._lock.release()
                    if rci:
                        args = req.kwargs['args']
                        kwargs = req.kwargs['kwargs']
                        try:
                            reply = Coro(rci._method, *args, **kwargs)
                        except:
                            reply = Exception(traceback.format_exc())
                    else:
                        reply = Exception('RCI "%s" is not registered' % req.kwargs['name'])
                yield conn.send_msg(serialize(reply))

            elif req.name == 'locate_coro':
                Coro._asyncoro._lock.acquire()
                coro = Coro._asyncoro._rcoros.get(req.kwargs['name'], None)
                Coro._asyncoro._lock.release()
                if not coro:
                    coro = self._rcoros.get(req.kwargs['name'], None)
                yield conn.send_msg(serialize(coro))

            elif req.name == 'locate_channel':
                Channel._asyncoro._lock.acquire()
                channel = Channel._asyncoro._rchannels.get('~' + req.kwargs['name'], None)
                Channel._asyncoro._lock.release()
                if not channel:
                    channel = self._rchannels.get('!' + req.kwargs['name'], None)
                yield conn.send_msg(serialize(channel))

            elif req.name == 'locate_rci':
                RCI._asyncoro._lock.acquire()
                rci = RCI._asyncoro._rcis.get(req.kwargs['name'], None)
                RCI._asyncoro._lock.release()
                yield conn.send_msg(serialize(rci))

            elif req.name == 'monitor':
                # synchronous message
                assert req.dst == self._location
                reply = -1
                monitor = req.kwargs.get('monitor', None)
                coro = req.kwargs.get('coro', None)
                name = req.kwargs.get('name', None)
                if coro and name:
                    if name[0] == '~':
                        Coro._asyncoro._lock.acquire()
                        coro = Coro._asyncoro._coros.get(int(coro), None)
                        if coro and coro._name == name:
                            reply = Coro._asyncoro._monitor(monitor, coro)
                        Coro._asyncoro._lock.release()
                    elif name == '!':
                        coro = self._coros.get(int(coro), None)
                        if coro and coro._name == name:
                            reply = self._monitor(monitor, coro)
                yield conn.send_msg(serialize(reply))

            elif req.name == 'terminate_coro':
                reply = -1
                coro = req.kwargs.get('coro', None)
                name = req.kwargs.get('name', None)
                if coro and name:
                    if name[0] == '~':
                        Coro._asyncoro._lock.acquire()
                        coro = Coro._asyncoro._coros.get(int(coro), None)
                        Coro._asyncoro._lock.release()
                    elif name[0] == '!':
                        coro = self._coros.get(int(coro), None)
                if isinstance(coro, Coro):
                    reply = coro.terminate()
                yield conn.send_msg(serialize(reply))

            elif req.name == 'subscribe':
                # synchronous message
                assert req.dst == self._location
                reply = -1
                channel = req.kwargs.get('channel', ' ')
                if channel[0] == '~':
                    Channel._asyncoro._lock.acquire()
                    channel = Channel._asyncoro._channels.get(channel, None)
                    Channel._asyncoro._lock.release()
                elif channel[0] == '!':
                    channel = self._channels.get(channel, None)
                if isinstance(channel, Channel) and channel._location == self._location:
                    subscriber = req.kwargs.get('subscriber', None)
                    if isinstance(subscriber, Coro):
                        if subscriber._location == self._location:
                            Coro._asyncoro._lock.acquire()
                            subscriber = Coro._asyncoro._coros.get(int(subscriber._id), None)
                            Coro._asyncoro._lock.release()
                        reply = yield channel.subscribe(subscriber)
                    elif isinstance(subsriber, Channel):
                        if subscriber._location == self._location:
                            Channel._asyncoro._lock.acquire()
                            subscriber = self._channels.get(subscriber._name, None)
                            Channel._asyncoro._lock.release()
                        reply = yield channel.subscribe(subscriber)
                yield conn.send_msg(serialize(reply))

            elif req.name == 'unsubscribe':
                # synchronous message
                assert req.dst == self._location
                reply = -1
                channel = req.kwargs.get('channel', ' ')
                if channel[0] == '~':
                    Channel._asyncoro._lock.acquire()
                    channel = Channel._asyncoro._channels.get(channel, None)
                    Channel._asyncoro._lock.release()
                elif channel[0] == '!':
                    channel = self._channels.get(channel, None)
                if isinstance(channel, Channel) and channel._location == self._location:
                    subscriber = req.kwargs.get('subscriber', None)
                    if isinstance(subscriber, Coro):
                        if subscriber._location == self._location:
                            Coro._asyncoro._lock.acquire()
                            subscriber = Coro._asyncoro._coros.get(int(subscriber._id), None)
                            Coro._asyncoro._lock.release()
                        reply = yield channel.unsubscribe(subscriber)
                    elif isinstance(subsriber, Channel):
                        if subscriber._location == self._location:
                            Channel._asyncoro._lock.acquire()
                            subscriber = self._channels.get(subscriber._name, None)
                            Channel._asyncoro._lock.release()
                        reply = yield channel.unsubscribe(subscriber)
                yield conn.send_msg(serialize(reply))

            elif req.name == 'locate_peer':
                if req.kwargs['name'] == self._name:
                    loc = self._location
                elif req.dst == self._location:
                    loc = None
                yield conn.send_msg(serialize(loc))

            elif req.name == 'send_file':
                # synchronous message
                assert req.dst == self._location
                sep = req.kwargs['sep']
                tgt = req.kwargs['file'].split(sep)[-1]
                if req.kwargs['dir']:
                    dir = os.path.join(*(req.kwargs['dir'].split(sep)))
                    if dir:
                        tgt = os.path.join(dir, tgt)
                tgt = os.path.abspath(os.path.join(self.__dest_path, tgt))
                stat_buf = req.kwargs['stat_buf']
                resp = 0
                if self.max_file_size and stat_buf.st_size > self.max_file_size:
                    logger.warning('file "%s" too big (%s) - must be smaller than %s',
                                   req.kwargs['file'], stat_buf.st_size, self.max_file_size)
                    resp = -1
                elif not tgt.startswith(self.__dest_path):
                    resp = -1
                elif os.path.isfile(tgt):
                    sbuf = os.stat(tgt)
                    if abs(stat_buf.st_mtime - sbuf.st_mtime) <= 1 and \
                       stat_buf.st_size == sbuf.st_size and \
                       stat.S_IMODE(stat_buf.st_mode) == stat.S_IMODE(sbuf.st_mode):
                        resp = stat_buf.st_size
                    elif not req.kwargs['overwrite']:
                        resp = -1

                if resp == 0:
                    try:
                        if not os.path.isdir(os.path.dirname(tgt)):
                            os.makedirs(os.path.dirname(tgt))
                        fd = open(tgt, 'wb')
                    except:
                        logger.debug('failed to create "%s" : %s', tgt, traceback.format_exc())
                        resp = -1
                if resp == 0:
                    recvd = 0
                    try:
                        while recvd < stat_buf.st_size:
                            yield conn.send_msg(serialize(recvd))
                            data = yield conn.recvall(min(stat_buf.st_size-recvd, 1024000))
                            if not data:
                                break
                            fd.write(data)
                            recvd += len(data)
                    except:
                        logger.warning('copying file "%s" failed', tgt)
                    finally:
                        fd.close()
                    if recvd == stat_buf.st_size:
                        os.utime(tgt, (stat_buf.st_atime, stat_buf.st_mtime))
                        os.chmod(tgt, stat.S_IMODE(stat_buf.st_mode))
                        resp = recvd
                    else:
                        os.remove(tgt)
                        resp = -1
                yield conn.send_msg(serialize(resp))

            elif req.name == 'del_file':
                # synchronous message
                assert req.dst == self._location
                tgt = os.path.basename(req.kwargs['file'])
                dir = req.kwargs['dir']
                if isinstance(dir, str) and dir:
                    tgt = os.path.join(dir, tgt)
                tgt = os.path.join(self.__dest_path, tgt)
                if tgt.startswith(self.__dest_path) and os.path.isfile(tgt):
                    os.remove(tgt)
                    d = os.path.dirname(tgt)
                    try:
                        while d > self.__dest_path and os.path.isdir(d):
                            os.rmdir(d)
                            d = os.path.dirname(d)
                    except:
                        # logger.debug(traceback.format_exc())
                        pass
                    reply = 0
                else:
                    reply = -1
                yield conn.send_msg(serialize(reply))

            elif req.name == 'peer':
                # synchronous message
                if req.kwargs.get('version', None) != __version__:
                    yield conn.send_msg(serialize(-1))
                    break
                auth = hashlib.sha1((req.kwargs['signature'] + self._secret).encode()).hexdigest()
                peer_loc = req.kwargs['from']
                yield conn.send_msg(serialize({'version': __version__, 'name': self._name}))
                _Peer(req.kwargs['name'], peer_loc, auth, self._keyfile, self._certfile)

            elif req.name == 'close_peer':
                # synchronous message
                peer_loc = req.kwargs.get('location', None)
                if peer_loc:
                    # TODO: remove from _stream_peers?
                    # _SysAsynCoro_._asyncoro._stream_peers.pop((peer_loc.addr, peer_loc.port))
                    _Peer.remove(peer_loc)
                yield conn.send_msg(b'closed')
                break

            elif req.name == 'acquaint':
                if req.kwargs.get('version', None) != __version__:
                    yield conn.send_msg(serialize(-1))
                    break
                peer_location = req.kwargs.get('location', None)
                if not isinstance(peer_location, Location) or peer_location == self._location:
                    yield conn.send_msg(serialize(-1))
                    break
                _Peer._lock.acquire()
                peer = _Peer.peers.get((peer_location.addr, peer_location.port), None)
                _Peer._lock.release()
                if self._secret is None:
                    auth_code = None
                else:
                    auth_code = hashlib.sha1((ping_info.get('signature', '') +
                                              self._secret).encode()).hexdigest()
                if peer and peer.auth != auth_code:
                    _Peer.remove(peer_location)
                    peer = None
                if not peer:
                    SysCoro(self._acquaint_, peer_location, req.kwargs['signature'])
                yield conn.send_msg(serialize(0))

            elif req.name == 'relay_ping':
                yield conn.send_msg(serialize(0))
                ping_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
                ping_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                ping_sock.settimeout(2)
                ping_msg = 'ping:'.encode() + serialize(req.kwargs)
                try:
                    yield ping_sock.sendto(ping_msg,
                                           (self._broadcast, self._udp_sock.getsockname()[1]))
                except:
                    pass
                finally:
                    ping_sock.close()

            else:
                logger.warning('invalid request "%s" ignored', req.name)

        conn.close()

    def _swing_call_(self, swing, method, *args, **kwargs):
        swing['result'] = yield method(*args, **kwargs)
        swing['event'].set()

    def __repr__(self):
        s = str(self._location)
        if s == self._name:
            return s
        else:
            return '"%s" @ %s' % (self._name, s)


asyncoro._NetRequest = _NetRequest
asyncoro._Peer = _Peer
asyncoro.SysCoro = SysCoro
asyncoro.AsynCoro = AsynCoro



"""
This file is part of asyncoro project.
See http://asyncoro.sourceforge.net for details.
"""

import sys
import os
import threading
import json
import cgi
import time
import socket
import ssl
import re
import traceback

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import DiscoroStatus, DiscoroNodeAvailInfo
import asyncoro.discoro as discoro

if sys.version_info.major > 2:
    import http.server as BaseHTTPServer
    from urllib.parse import urlparse
else:
    import BaseHTTPServer
    from urlparse import urlparse


__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__email__ = "pgiri@yahoo.com"
__copyright__ = "Copyright 2015, Giridhar Pemmasani"
__contributors__ = []
__maintainer__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"

__all__ = ['HTTPServer']


if sys.version_info.major >= 3:
    def dict_iter(arg, iterator):
        return getattr(arg, iterator)()
else:
    def dict_iter(arg, iterator):
        return getattr(arg, 'iter' + iterator)()


class HTTPServer(object):

    NodeStatus = {discoro.Scheduler.NodeDiscovered: 'Discovered',
                  discoro.Scheduler.NodeInitialized: 'Initialized',
                  discoro.Scheduler.NodeClosed: 'Closed',
                  discoro.Scheduler.NodeIgnore: 'Ignore',
                  discoro.Scheduler.NodeDisconnected: 'Disconnected'}
    ServerStatus = {discoro.Scheduler.ServerDiscovered: 'Discovered',
                    discoro.Scheduler.ServerInitialized: 'Initialized',
                    discoro.Scheduler.ServerClosed: 'Closed',
                    discoro.Scheduler.ServerIgnore: 'Ignore',
                    discoro.Scheduler.ServerDisconnected: 'Disconnected'}

    class _Node(object):
        def __init__(self, name, addr):
            self.name = name
            self.addr = addr
            self.status = None
            self.servers = {}
            self.update_time = time.time()
            self.coros_submitted = 0
            self.coros_done = 0
            self.avail_info = None

    class _Server(object):
        def __init__(self, location):
            self.location = location
            self.status = None
            self.coros = {}
            self.coros_submitted = 0
            self.coros_done = 0

    class _HTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

        def __init__(self, ctx, DocumentRoot, *args):
            self._ctx = ctx
            self.DocumentRoot = DocumentRoot
            BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, *args)

        def log_message(self, fmt, *args):
            # asyncoro.logger.debug('HTTP client %s: %s', self.client_address[0], fmt % args)
            return

        @staticmethod
        def json_encode_nodes(arg):
            nodes = [dict(node.__dict__) for node in dict_iter(arg, 'values')]
            for node in nodes:
                node['servers'] = len(node['servers'])
                node['avail_info'] = node['avail_info'].__dict__
            return nodes

        def do_GET(self):
            if self.path == '/cluster_updates':
                self._ctx._lock.acquire()
                nodes = self.__class__.json_encode_nodes(self._ctx._updates)
                self._ctx._updates.clear()
                self._ctx._lock.release()
                self.send_response(200)
                self.send_header('Content-Type', 'application/json; charset=utf-8')
                self.end_headers()
                self.wfile.write(json.dumps(nodes).encode())
                return
            elif self.path == '/cluster_status':
                self._ctx._lock.acquire()
                nodes = self.__class__.json_encode_nodes(self._ctx._nodes)
                self._ctx._lock.release()
                self.send_response(200)
                self.send_header('Content-Type', 'application/json; charset=utf-8')
                self.end_headers()
                self.wfile.write(json.dumps(nodes).encode())
                return
            else:
                parsed_path = urlparse(self.path)
                path = parsed_path.path.lstrip('/')
                if not path or path == 'index.html':
                    path = 'cluster.html'
                path = os.path.join(self.DocumentRoot, path)
                try:
                    with open(path) as fd:
                        data = fd.read()
                    if path.endswith('.html'):
                        if path.endswith('.html'):
                            data = data % {'TIMEOUT': str(self._ctx._poll_sec)}
                        content_type = 'text/html'
                    elif path.endswith('.js'):
                        content_type = 'text/javascript'
                    elif path.endswith('.css'):
                        content_type = 'text/css'
                    elif path.endswith('.ico'):
                        content_type = 'image/x-icon'
                    self.send_response(200)
                    self.send_header('Content-Type', content_type)
                    if content_type == 'text/css' or content_type == 'text/javascript':
                        self.send_header('Cache-Control', 'private, max-age=86400')
                    self.end_headers()
                    self.wfile.write(data.encode())
                    return
                except:
                    asyncoro.logger.warning('HTTP client %s: Could not read/send "%s"',
                                            self.client_address[0], path)
                    asyncoro.logger.debug(traceback.format_exc())
                self.send_error(404)
                return
            asyncoro.logger.debug('Bad GET request from %s: %s',
                                  self.client_address[0], self.path)
            self.send_error(400)
            return

        def do_POST(self):
            form = cgi.FieldStorage(fp=self.rfile, headers=self.headers,
                                    environ={'REQUEST_METHOD': 'POST'})
            if self.path == '/server_info':
                server = None
                max_coros = 0
                for item in form.list:
                    if item.name == 'location':
                        m = re.match('^(\d+[\.\d]+):(\d+)$', item.value)
                        if m:
                            node = self._ctx._nodes.get(m.group(1))
                            if node:
                                server = node.servers.get(m.group(0))
                    elif item.name == 'limit':
                        try:
                            max_coros = int(item.value)
                        except:
                            pass
                if server:
                    if 0 < max_coros < len(server.coros):
                        rcoros = []
                        for i, rcoro in enumerate(dict_iter(server.coros, 'values')):
                            if i >= max_coros:
                                break
                            rcoros.append(rcoro)
                    else:
                        rcoros = server.coros.values()
                    rcoros = [{'coro': str(rcoro.coro), 'name': rcoro.coro.name,
                               'args': ', '.join(str(arg) for arg in rcoro.args),
                               'kwargs': ', '.join('%s=%s' % (key, val)
                                                   for key, val in rcoro.kwargs.items()),
                               'start_time': rcoro.start_time
                               } for rcoro in rcoros]
                    info = {'location': str(server.location),
                            'status': server.status, 'coros_submitted': server.coros_submitted,
                            'coros_done': server.coros_done, 'coros': rcoros,
                            'update_time': node.update_time}
                else:
                    info = {}
                self.send_response(200)
                self.send_header('Content-Type', 'application/json; charset=utf-8')
                self.end_headers()
                self.wfile.write(json.dumps(info).encode())
                return
            elif self.path == '/node_info':
                addr = None
                for item in form.list:
                    if item.name == 'host':
                        if re.match('^(\d+[\.\d]+)$', item.value):
                            addr = item.value
                        else:
                            try:
                                addr = socket.gethostbyname(item.value)
                            except:
                                addr = item.value
                        break
                node = self._ctx._nodes.get(addr)
                if node:
                    info = {'addr': node.addr, 'name': node.name,
                            'status': node.status, 'update_time': node.update_time,
                            'avail_info': node.avail_info.__dict__,
                            'coros_submitted': node.coros_submitted, 'coros_done': node.coros_done,
                            'servers': [
                                {'location': str(server.location),
                                 'coros_submitted': server.coros_submitted,
                                 'coros_done': server.coros_done,
                                 'coros_running': len(server.coros),
                                 'update_time': node.update_time
                                 } for server in node.servers.values()
                                ]
                            }
                else:
                    info = {}
                self.send_response(200)
                self.send_header('Content-Type', 'application/json; charset=utf-8')
                self.end_headers()
                self.wfile.write(json.dumps(info).encode())
                return
            elif self.path == '/terminate_coros':
                coros = []
                for item in form.list:
                    if item.name == 'coro':
                        try:
                            coros.append(item.value)
                        except ValueError:
                            asyncoro.logger.debug('Terminate: coro "%s" is invalid', item.value)

                terminated = []
                self._ctx._lock.acquire()
                for coro in coros:
                    s = coro.split('@')
                    if len(s) != 2:
                        continue
                    location = s[1]
                    s = location.split(':')
                    if len(s) != 2:
                        continue
                    node = self._ctx._nodes.get(s[0])
                    if not node:
                        continue
                    server = node.servers.get(location)
                    if not server:
                        continue
                    rcoro = server.coros.get(coro)
                    if rcoro and rcoro.coro.terminate() == 0:
                        terminated.append(coro)
                self._ctx._lock.release()
                self.send_response(200)
                self.send_header('Content-Type', 'application/json; charset=utf-8')
                self.end_headers()
                self.wfile.write(json.dumps(terminated).encode())
                return

            elif self.path == '/set_poll_sec':
                for item in form.list:
                    if item.name != 'timeout':
                        continue
                    try:
                        timeout = int(item.value)
                        if timeout < 1:
                            timeout = 0
                    except:
                        asyncoro.logger.warning('HTTP client %s: invalid timeout "%s" ignored',
                                                self.client_address[0], item.value)
                        timeout = 0
                    self._ctx._poll_sec = timeout
                    self.send_response(200)
                    self.send_header('Content-Type', 'text/html')
                    self.end_headers()
                    return
            asyncoro.logger.debug('Bad POST request from %s: %s',
                                  self.client_address[0], self.path)
            self.send_error(400)
            return

    def __init__(self, computation, host='', port=8181, poll_sec=10, DocumentRoot=None,
                 keyfile=None, certfile=None):
        self._lock = threading.Lock()
        if not DocumentRoot:
            DocumentRoot = os.path.join(os.path.dirname(__file__), 'data')
        self._nodes = {}
        self._updates = {}
        if poll_sec < 1:
            asyncoro.logger.warning('invalid poll_sec value %s; it must be at least 1', poll_sec)
            poll_sec = 1
        self._poll_sec = poll_sec
        self._server = BaseHTTPServer.HTTPServer((host, port), lambda *args:
                                  HTTPServer._HTTPRequestHandler(self, DocumentRoot, *args))
        if certfile:
            self._server.socket = ssl.wrap_socket(self._server.socket, keyfile=keyfile,
                                                  certfile=certfile, server_side=True)
        self._httpd_thread = threading.Thread(target=self._server.serve_forever)
        self._httpd_thread.daemon = True
        self._httpd_thread.start()
        self.status_coro = asyncoro.SysCoro(self.status_proc)
        self.computation = computation
        if not computation.status_coro:
            computation.status_coro = self.status_coro
        asyncoro.logger.info('Started HTTP%s server at %s',
                             's' if certfile else '', str(self._server.socket.getsockname()))

    def status_proc(self, coro=None):
        coro.set_daemon()
        while True:
            msg = yield coro.receive()
            if isinstance(msg, asyncoro.MonitorException):
                rcoro = msg.args[0]
                node = self._nodes.get(rcoro.location.addr)
                if node:
                    server = node.servers.get(str(rcoro.location))
                    if server:
                        if server.coros.pop(str(rcoro), None) is not None:
                            server.coros_done += 1
                            node.coros_done += 1
                            node.update_time = time.time()
                            self._updates[node.addr] = node
            elif isinstance(msg, DiscoroStatus):
                if msg.status == discoro.Scheduler.CoroCreated:
                    rcoro = msg.info
                    node = self._nodes.get(rcoro.coro.location.addr)
                    if node:
                        server = node.servers.get(str(rcoro.coro.location))
                        if server:
                            server.coros[str(rcoro.coro)] = rcoro
                            server.coros_submitted += 1
                            node.coros_submitted += 1
                            node.update_time = time.time()
                            self._updates[node.addr] = node
                elif msg.status == discoro.Scheduler.ServerInitialized:
                    node = self._nodes.get(msg.info.addr)
                    if node:
                        server = node.servers.get(str(msg.info), None)
                        if not server:
                            server = HTTPServer._Server(msg.info)
                            node.servers[str(server.location)] = server
                        server.status = HTTPServer.ServerStatus[msg.status]
                        node.update_time = time.time()
                        self._updates[node.addr] = node
                elif (msg.status == discoro.Scheduler.ServerClosed or
                      msg.status == discoro.Scheduler.ServerDisconnected):
                    node = self._nodes.get(msg.info.addr)
                    if node:
                        server = node.servers.get(str(msg.info), None)
                        if server:
                            server.status = HTTPServer.ServerStatus[msg.status]
                            node.update_time = time.time()
                            self._updates[node.addr] = node
                elif msg.status == discoro.Scheduler.NodeInitialized:
                    node = self._nodes.get(msg.info.addr)
                    if not node:
                        node = HTTPServer._Node(msg.info.name, msg.info.addr)
                        node.avail_info = msg.info.avail_info
                        node.avail_info.location = None
                        self._nodes[msg.info.addr] = node
                    node.status = HTTPServer.NodeStatus[msg.status]
                    node.update_time = time.time()
                    self._updates[node.addr] = node
                elif (msg.status == discoro.Scheduler.NodeClosed or
                      msg.status == discoro.Scheduler.NodeDisconnected):
                    node = self._nodes.get(msg.info.addr)
                    if node:
                        node.status = HTTPServer.NodeStatus[msg.status]
                        node.update_time = time.time()
                        self._updates[node.addr] = node
            elif isinstance(msg, DiscoroNodeAvailInfo):
                node = self._nodes.get(msg.location.addr, None)
                if node:
                    node.avail_info = msg
                    node.avail_info.location = None
                    node.update_time = time.time()
                    self._updates[node.addr] = node
            else:
                asyncoro.logger.warning('Status message ignored: %s', type(msg))

    def shutdown(self, wait=True):
        """This method should be called by user program to close the
        http server. If 'wait' is True the server waits for poll_sec
        so the http client gets all the updates before server is
        closed.
        """
        if wait:
            asyncoro.logger.info('HTTP server waiting for %s seconds for client updates '
                                 'before quitting', self._poll_sec)
            if asyncoro.AsynCoro().cur_coro():
                def _shutdown(coro=None):
                    yield coro.sleep(self._poll_sec + 0.5)
                    self._server.shutdown()
                    self._server.server_close()
                asyncoro.Coro(_shutdown)
            else:
                time.sleep(self._poll_sec + 0.5)
                self._server.shutdown()
                self._server.server_close()
        else:
            self._server.shutdown()
            self._server.server_close()



#!/usr/bin/python3

"""This file is part of asyncoro; see http://asyncoro.sourceforge.net
for details.

This module provides API for creating distributed communicating
processes. 'Computation' class should be used to package computation
components (Python generator functions, Python functions, files,
classes, modules) and then schedule runs that create remote coroutines
at remote server processes running 'discoronode.py'.

See 'discoro_client*.py' files in 'examples' directory for various use
cases.
"""

import os
import sys
import inspect
import hashlib
import collections
import time
import socket
import shutil
import operator
import functools
import re

import asyncoro.disasyncoro as asyncoro
from asyncoro import Coro, SysCoro, logger

__author__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
__copyright__ = "Copyright (c) 2014-2015 Giridhar Pemmasani"
__license__ = "MIT"
__url__ = "http://asyncoro.sourceforge.net"

__all__ = ['Scheduler', 'Computation', 'DiscoroStatus', 'DiscoroCoroInfo',
           'DiscoroNodeInfo', 'DiscoroNodeAvailInfo', 'DiscoroNodeAllocate']

MsgTimeout = asyncoro.MsgTimeout
MinPulseInterval = 10
MaxPulseInterval = 10 * MinPulseInterval

# status about nodes / servers are sent with this structure
DiscoroStatus = collections.namedtuple('DiscoroStatus', ['status', 'info'])
DiscoroCoroInfo = collections.namedtuple('DiscoroCoroInfo', ['coro', 'args', 'kwargs', 'start_time'])
DiscoroNodeInfo = collections.namedtuple('DiscoroNodeInfo', ['name', 'addr', 'cpus', 'platform',
                                                             'avail_info'])
# for internal use only
_DiscoroFunction = collections.namedtuple('_DiscoroFunction', ['name', 'code', 'args', 'kwargs'])


class DiscoroNodeAvailInfo(object):
    """Node availability status is indicated with this class.  'cpu' is
    available CPU in percent in the range 0 to 100. 0 indicates node is busy
    executing tasks on all CPUs and 100 indicates node is not busy at all.
    """

    def __init__(self, location, cpu, memory, disk, swap):
        self.location = location
        self.cpu = cpu
        self.memory = memory
        self.disk = disk
        self.swap = swap


class DiscoroNodeAllocate(object):
    """Allocation of nodes can be customized by specifying 'node_allocations' of
    Computation with DiscoroNodeAllocate instances.

    'ip_addr' must be a regular expression, e.g., "192\.168\.1\..*" (note that
    "." without escape stands for "any" character), 'cpus' must be minimum
    number of servers running on that node, 'memory' is available memory in
    bytes, 'disk' is available disk space (on the partition where discoronode
    servers are running from), and 'platform' is regular expression to match
    output of"platform.platform()" on that node, e.g., "Linux.*x86_64" to accept
    only nodes that run 64-bit Linux.
    """

    def __init__(self, node, platform='', cpus=0, memory=0, disk=0):
        if node.find('*') < 0:
            try:
                node = socket.gethostbyname(node)
            except:
                node = ''

        if node:
            self.ip_rex = node.replace('.', '\\.').replace('*', '.*')
        else:
            logger.warning('node "%s" is invalid', node)
            self.ip_rex = ''
        self.platform = platform.lower()
        self.cpus = cpus
        self.memory = memory
        self.disk = disk

    def allocate(self, ip_addr, name, platform, cpus, memory, disk):
        """When a node is found, scheduler calls this method with IP address,
        name, CPUs, memory and disk available on that node. This method should
        return a number indicating number of CPUs to use. If return value is 0,
        the node is not used; if the return value is < 0, this allocation is
        ignored (next allocation in the 'node_allocations' list, if any, is
        applied).
        """
        if not re.match(self.ip_rex, ip_addr):
            return -1
        if (self.platform and not re.search(self.platform, platform)):
            return -1
        if ((self.memory and memory and self.memory > memory) or
            (self.disk and disk and self.disk > disk)):
            return 0
        if self.cpus > 0:
            if self.cpus > cpus:
                return 0
            return self.cpus
        elif self.cpus == 0:
            return 0
        else:
            cpus += self.cpus
            if cpus < 0:
                return 0
            return cpus


class Computation(object):
    """Packages components to distribute to remote asyncoro schedulers to create
    (remote) coroutines.
    """

    def __init__(self, components, status_coro=None, timeout=MsgTimeout,
                 pulse_interval=(2*MinPulseInterval), ping_interval=None, zombie_period=0,
                 node_allocations=[], node_available=None, node_setup=None,
                 peers_communicate=False):
        """'components' should be a list, each element of which is either a
        module, a (generator or normal) function, path name of a file, a class
        or an object (in which case the code for its class is sent).

        'status_coro', if not None, should be a coroutine. The scheduler sends
        status messages indicating when a remote server process has been
        initialized (so it is ready to run jobs), closed etc., and exit status
        of remote coroutines. See 'discoro_client*.py' files in examples
        directory.

        'timeout' is maximum number of seconds to complete a communication
        (transfer of messages). If client / scheduler / remote servers couldn't
        send / receive a message within this period, the operation is
        aborted. Bigger values may be used when communicating over slower
        connections.

        'pulse_interval' is interval (number of seconds) used for heart beat
        messages to check if client / scheduler / server is alive. If the other
        side doesn't reply to 5 heart beat messages, it is treated as dead.

        'zombie_period' is maximum number of seconds a server process stays idle
        (i.e., no coroutines running on that server) before the computation is
        automatically closed (on that server). Once closed, the computation
        can't use that server anymore. This discards unused clients so other
        pending (queued) computations can use compute resources. If
        'zombie_period' is 0, the servers don't check for idle period and
        don't close computation (until the user program explicitly closes
        it).
        """

        if status_coro is not None and not isinstance(status_coro, Coro):
            raise Exception('"status_coro" must be coroutine')
        if timeout < 1 or timeout > MaxPulseInterval:
            raise Exception('"timeout" must be at least 1 and at most %s' % MaxPulseInterval)
        if pulse_interval < MinPulseInterval or pulse_interval > MaxPulseInterval:
            raise Exception('"pulse_interval" must be at least %s and at most %s' %
                            (MinPulseInterval, MaxPulseInterval))
        if ping_interval and ping_interval < MinPulseInterval:
            raise Exception('"ping_interval" must be at least %s', MinPulseInterval)
        if (not isinstance(zombie_period, (int, float)) or 0 > zombie_period < MaxPulseInterval):
            raise Exception('"zombie_period" must be either 0 or >= %s' % MaxPulseInterval)
        if ((not isinstance(node_allocations, list)) or
            any(not isinstance(_, (DiscoroNodeAllocate, str)) for _ in node_allocations)):
            raise Exception('"node_allocations" must be list of DiscoroNodeAllocate instances')
        if node_available and not inspect.isgeneratorfunction(node_available):
            raise Exception('"node_available" must be a coroutine (generator function)')
        if node_setup and not inspect.isgeneratorfunction(node_setup):
            raise Exception('"node_setup" must be a coroutine (generator function)')

        if not isinstance(components, list):
            components = [components]

        self._code = ''
        self._xfer_funcs = set()
        self._xfer_files = []
        self.status_coro = status_coro
        self._auth = None
        self.scheduler = None
        self._pulse_coro = None
        self._pulse_interval = pulse_interval
        self._ping_interval = ping_interval
        self.timeout = timeout
        self.zombie_period = zombie_period
        self._node_allocations = [node if isinstance(node, DiscoroNodeAllocate)
                                  else DiscoroNodeAllocate(node) for node in node_allocations]
        self._node_available = node_available
        if node_setup:
            components.append(node_setup)
            self._node_setup = node_setup.__name__
        else:
            self._node_setup = None
        self._peers_communicate = bool(peers_communicate)
        depends = set()
        cwd = os.getcwd()
        for dep in components:
            if isinstance(dep, str) or inspect.ismodule(dep):
                if inspect.ismodule(dep):
                    name = dep.__file__
                    if name.endswith('.pyc'):
                        name = name[:-1]
                    if not name.endswith('.py'):
                        raise Exception('Invalid module "%s" - must be python source.' % dep)
                    if name.startswith(cwd):
                        dst = os.path.dirname(name[len(cwd+os.sep):])
                    elif dep.__package__:
                        dst = dep.__package__.replace('.', os.sep)
                    else:
                        dst = os.path.dirname(dep.__name__.replace('.', os.sep))
                else:
                    name = os.path.abspath(dep)
                    if name.startswith(cwd):
                        dst = os.path.dirname(name[len(cwd+os.sep):])
                    else:
                        dst = '.'
                if name in depends:
                    continue
                try:
                    with open(name, 'rb') as fd:
                        pass
                except:
                    raise Exception('File "%s" is not valid' % name)
                self._xfer_files.append((name, dst, os.sep))
                depends.add(name)
            elif (inspect.isgeneratorfunction(dep) or inspect.isfunction(dep) or
                  inspect.isclass(dep) or hasattr(dep, '__class__')):
                if inspect.isgeneratorfunction(dep) or inspect.isfunction(dep):
                    name = dep.__name__
                elif inspect.isclass(dep):
                    name = dep.__name__
                elif hasattr(dep, '__class__') and inspect.isclass(dep.__class__):
                    dep = dep.__class__
                    name = dep.__name__
                if name in depends:
                    continue
                depends.add(name)
                self._xfer_funcs.add(name)
                self._code += '\n' + inspect.getsource(dep).lstrip()
            else:
                raise Exception('Invalid computation: %s' % dep)
        # check code can be compiled
        compile(self._code, '<string>', 'exec')
        # Under Windows discoro server may send objects with '__mp_main__'
        # scope, so make an alias to '__main__'.  Do so even if scheduler is not
        # running on Windows; it is possible the client is not Windows, but a
        # node is.
        if '__mp_main__' not in sys.modules:
            sys.modules['__mp_main__'] = sys.modules['__main__']

    def schedule(self, location=None, timeout=None):
        """Schedule computation for execution. Must be used with 'yield' as
        'result = yield compute.schedule()'. If scheduler is executing other
        computations, this will block until scheduler processes them
        (computations are processed in the order submitted).
        """

        if self._auth is not None:
            raise StopIteration(0)
        self._auth = ''
        if self.status_coro is not None and not isinstance(self.status_coro, Coro):
            raise StopIteration(-1)

        self.scheduler = yield Coro.locate('discoro_scheduler', location=location,
                                           timeout=self.timeout)
        if not isinstance(self.scheduler, Coro):
            raise StopIteration(-1)

        def _schedule(self, coro=None):
            self._pulse_coro = SysCoro(self._pulse_proc)
            if self._node_available and self.scheduler.location != self._pulse_coro.location:
                logger.warning('"node_available" is disabled with shared scheduler')
                self._node_available = True
            msg = {'req': 'schedule', 'computation': asyncoro.serialize(self), 'client': coro}
            self.scheduler.send(msg)
            self._auth = yield coro.receive(timeout=self.timeout)
            if not isinstance(self._auth, str):
                logger.debug('Could not send computation to scheduler %s: %s',
                             self.scheduler, self._auth)
                raise StopIteration(-1)
            SysCoro.scheduler().atexit(10, lambda: SysCoro(self.close).value())
            if coro.location != self.scheduler.location:
                for xf, dst, sep in self._xfer_files:
                    drive, xf = os.path.splitdrive(xf)
                    if xf.startswith(sep):
                        xf = os.path.join(os.sep, *(xf.split(sep)))
                    else:
                        xf = os.path.join(*(xf.split(sep)))
                    xf = drive + xf
                    dst = os.path.join(self._auth, os.path.join(*(dst.split(sep))))
                    if (yield asyncoro.AsynCoro.instance().send_file(
                       self.scheduler.location, xf, dir=dst, timeout=self.timeout)) < 0:
                        logger.warning('Could not send file "%s" to scheduler', xf)
                        yield self.close()
                        raise StopIteration(-1)
            msg = {'req': 'await', 'auth': self._auth, 'client': coro}
            self.scheduler.send(msg)
            resp = yield coro.receive(timeout=timeout)
            if (isinstance(resp, dict) and resp.get('auth') == self._auth and
               resp.get('resp') == 'scheduled'):
                raise StopIteration(0)
            else:
                yield self.close()
                raise StopIteration(-1)

        yield Coro(_schedule, self).finish()

    def run_at(self, where, gen, *args, **kwargs):
        """Run given generator function 'gen' with arguments 'args' and 'kwargs'
        at 'where'.  If the request is successful, 'rcoro' will be a (remote)
        coroutine; check result with 'isinstance(rcoro, asyncoro.Coro)'. Must be
        used with 'yield' as 'rcoro = yield compute.run_at(loc, genf, ...)'.

        If 'where' is a string, it is assumed to be IP address of a node, in
        which case the coroutine is scheduled at that node on a server with
        least load (i.e., server with least number of pending coroutines). If
        'where' is a Location instance, it is assumed to be server location in
        which case the coroutine is scheduled at that server.

        'gen' must be generator function, as it is used to run coroutine at
        remote location.
        """
        if isinstance(gen, str):
            name = gen
        else:
            name = gen.__name__

        if name in self._xfer_funcs:
            code = None
        else:
            # if not inspect.isgeneratorfunction(gen):
            #     logger.warning('"%s" is not a valid generator function', name)
            #     raise StopIteration(None)
            code = inspect.getsource(gen).lstrip()

        def _run(self, coro=None):
            msg = {'req': 'run', 'auth': self._auth, 'where': where, 'client': coro,
                   'func': asyncoro.serialize(_DiscoroFunction(name, code, args, kwargs))}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) == 1:
                rcoro = yield coro.receive(self.timeout)
                if self.status_coro and isinstance(rcoro, Coro):
                    msg = DiscoroCoroInfo(rcoro, args, kwargs, time.time())
                    self.status_coro.send(DiscoroStatus(Scheduler.CoroCreated, msg))
            else:
                rcoro = None
            raise StopIteration(rcoro)

        yield Coro(_run, self).finish()

    def run_each(self, where, gen, *args, **kwargs):
        """Run given generator function 'gen' with arguments 'args' and 'kwargs'
        at each node or server.  If the request is successful, 'rcoro' will be a
        (remote) coroutine; check result with 'isinstance(rcoro,
        asyncoro.Coro)'. Must be used with 'yield' as 'rcoros = yield
        compute.run_each(loc, genf, ...)'.

        If 'where' is 'node', the function is scheduled at every node, if
        'where' is 'server', the function is scheduled at every server (on every
        node); otherwise, 'where' must be IP address, in which case the function
        is scheduled at every server at that node.
        """
        if isinstance(gen, str):
            name = gen
        else:
            name = gen.__name__

        if name in self._xfer_funcs:
            code = None
        else:
            # if not inspect.isgeneratorfunction(gen):
            #     logger.warning('"%s" is not a valid generator function', name)
            #     raise StopIteration([])
            code = inspect.getsource(gen).lstrip()

        def _run(self, coro=None):
            msg = {'req': 'run_each', 'auth': self._auth, 'where': where, 'client': coro,
                   'func': asyncoro.serialize(_DiscoroFunction(name, code, args, kwargs))}
            # TODO: timeout should be for all operations combined?
            n = yield self.scheduler.deliver(msg, timeout=self.timeout)
            if n != 1:
                raise StopIteration([])
            n = yield coro.receive(timeout=self.timeout)
            rcoros = []
            for i in range(n):
                rcoro = yield coro.receive(timeout=self.timeout)
                if isinstance(rcoro, Coro):
                    rcoros.append(rcoro)
                    if self.status_coro:
                        msg = DiscoroCoroInfo(rcoro, args, kwargs, time.time())
                        self.status_coro.send(DiscoroStatus(Scheduler.CoroCreated, msg))
            raise StopIteration(rcoros)

        yield Coro(_run, self).finish()

    def run(self, gen, *args, **kwargs):
        """Run given generator function 'gen' with arguments 'args' and 'kwargs'
        at a server with least load at a node with least load.

        Must be used with 'yield' as 'rcoro = yield compute.run(genf, ...)'. If
        the request is successful, 'rcoro' will be a (remote) coroutine.
        """
        yield self.run_at(None, gen, *args, **kwargs)

    def run_nodes(self, gen, *args, **kwargs):
        """Run given generator function 'gen' with arguments 'args' and 'kwargs'
        at a server with least load at every node.

        Must be used with 'yield' as 'rcoros = yield compute.run_nodes(genf,
        ...)'. 'rcoros' will be a list of (remote) coroutines.
        """
        yield self.run_each('node', gen, *args, **kwargs)

    def run_servers(self, gen, *args, **kwargs):
        """Run given generator function 'gen' with arguments 'args' and 'kwargs'
        at every server (at every node).

        Must be used with 'yield' as 'rcoros = yield compute.run_servers(genf,
        ...)'. 'rcoros' will be a list of (remote) coroutines.
        """
        yield self.run_each('server', gen, *args, **kwargs)

    def run_node_servers(self, host, gen, *args, **kwargs):
        """Run given generator function 'gen' with arguments 'args' and 'kwargs'
        at every server at given node at 'host'. 'host' must IP address or host
        name of that node.

        Must be used with 'yield' as 'rcoros = yield
        compute.run_node_servers(host, genf, ...)'. 'rcoros' will be a list of
        (remote) coroutines scheduled at servers on that node.
        """
        if isinstance(host, str) and len(host) > 0:
            # if host starts with digit, assume IP address
            if not host[0].isdigit():
                host = socket.gethostbyname(host)
            yield self.run_each(host, gen, *args, **kwargs)
        else:
            raise StopIteration([])

    # TODO: add 'map' methods to run with arguments as iterators
    # (e.g., list of tuples)

    def nodes(self):
        """Get list of addresses of nodes initialized for this computation. Must
        be used with 'yield' as 'yield compute.nodes()'.
        """

        def _nodes_list(self, coro=None):
            msg = {'req': 'nodes_list', 'auth': self._auth, 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) == 1:
                yield coro.receive(self.timeout)
            else:
                raise StopIteration([])

        yield Coro(_nodes_list, self).finish()

    def servers(self):
        """Get list of Location instances of servers initialized for this
        computation. Must be used with 'yield' as 'yield compute.servers()'.
        """

        def _servers_list(self, coro=None):
            msg = {'req': 'servers_list', 'auth': self._auth, 'client': coro}
            if (yield self.scheduler.deliver(msg, timeout=self.timeout)) == 1:
                yield coro.receive(self.timeout)
            else:
                raise StopIteration([])

        yield Coro(_servers_list, self).finish()

    def close(self):
        """Close computation. Must be used with 'yield' as 'yield
        compute.close()'.
        """

        def _close(self, done, coro=None):
            msg = {'req': 'close_computation', 'auth': self._auth, 'client': coro}
            yield self.scheduler.deliver(msg, timeout=self.timeout)
            msg = yield coro.receive(timeout=self.timeout)
            if msg != 'closed':
                logger.warning('%s: closing computation failed?', self._auth)
            self._auth = None
            if self._pulse_coro:
                yield self._pulse_coro.send('quit')
                self._pulse_coro = None
            done.set()

        if self._auth:
            done = asyncoro.Event()
            SysCoro(_close, self, done)
            yield done.wait()

    def _pulse_proc(self, coro=None):
        """For internal use only.
        """
        coro.set_daemon()
        last_pulse = time.time()
        while 1:
            msg = yield coro.receive(timeout=(2 * self._pulse_interval))
            if msg == 'pulse':
                last_pulse = time.time()
            elif msg == 'quit':
                break
            elif msg is None:
                if self.zombie_period and (time.time() - last_pulse) > self.zombie_period:
                    logger.warning('scheduler is zombie!')
                    if self._auth:
                        yield self.close()
                    break
            else:
                logger.debug('ignoring invalid pulse message')
        self._pulse_coro = None


class Scheduler(object, metaclass=asyncoro.Singleton):

    # status indications ('status' attribute of DiscoroStatus)
    NodeDiscovered = 1
    NodeInitialized = 2
    NodeClosed = 3
    NodeIgnore = 4
    NodeDisconnected = 5

    ServerDiscovered = 11
    ServerInitialized = 12
    ServerClosed = 13
    ServerIgnore = 14
    ServerDisconnected = 15

    CoroCreated = 20
    ComputationScheduled = 23
    ComputationClosed = 25

    _instance = None

    """This class is for use by Computation class (see below) only.  Other than
    the status indications above, none of its attributes are to be accessed
    directly.
    """

    __status_coro = None

    class _Node(object):

        def __init__(self, name, addr):
            self.name = name
            self.addr = addr
            self.cpus = 0
            self.platform = None
            self.avail_info = None
            self.servers = {}
            self.ncoros = 0
            self.load = 0.0
            self.status = None
            self.coro = None

        def run(self, func, computation, client):
            where = None
            load = None
            for server in self.servers.values():
                if server.status != Scheduler.ServerInitialized:
                    continue
                if load is None or len(server.rcoros) < load:
                    where = server
                    load = len(server.rcoros)
            if where:
                yield where.run(func, computation, self, client)
            else:
                client.send(None)

    class _Server(object):

        def __init__(self, name, location):
            self.name = name
            self.location = location
            self.coro = None
            self.status = None
            self.rcoros = {}
            self.xfer_files = []
            self.last_pulse = time.time()
            self.askew_results = {}

        def run(self, func, computation, node, client):
            def _run(self, func, coro=None):
                self.coro.send({'req': 'run', 'auth': computation._auth, 'func': func,
                                'client': coro})
                rcoro = yield coro.receive(timeout=computation.timeout)
                if isinstance(rcoro, Coro):
                    # TODO: keep func too for fault-tolerance
                    self.rcoros[rcoro] = rcoro
                    if self.askew_results:
                        msg = self.askew_results.pop(rcoro, None)
                        if msg:
                            Scheduler.__status_coro.send(msg)
                        else:
                            node.ncoros += 1
                            node.load = float(node.ncoros) / len(node.servers)
                    else:
                        node.ncoros += 1
                        node.load = float(node.ncoros) / len(node.servers)
                else:
                    logger.debug('failed to create rcoro: %s / %s',
                                 str(rcoro), computation.timeout)
                raise StopIteration(rcoro)

            rcoro = yield SysCoro(_run, self, func).finish()
            yield client.deliver(rcoro)

    def __init__(self, **kwargs):
        self.__class__._instance = self
        self._nodes = {}
        self._cur_computation = None
        self.__cur_client_auth = None
        self.__cur_node_allocations = []
        self.__cur_node_available = None
        self.__pulse_interval = MinPulseInterval
        self.__ping_interval = None
        self.__sched_event = asyncoro.Event()
        self.__terminate = False
        self._shared = False
        self._node_port = kwargs.pop('discoronode_port', 51351)
        self.__server_locations = set()

        kwargs['name'] = 'discoro_scheduler'
        clean = kwargs.pop('clean', False)
        self.__zombie_period = kwargs.pop('zombie_period', None)
        nodes = kwargs.pop('nodes', [])
        self.asyncoro = asyncoro.AsynCoro.instance(**kwargs)
        self.__dest_path = os.path.join(self.asyncoro.dest_path, 'discoro', 'discoroscheduler')
        if clean:
            shutil.rmtree(self.__dest_path)
        self.asyncoro.dest_path = self.__dest_path

        self.__scheduler_coro = SysCoro(self.__scheduler_proc, nodes)
        self.__client_coro = SysCoro(self.__client_proc)
        self.__timer_coro = SysCoro(self.__timer_proc)
        Scheduler.__status_coro = self.__status_coro = SysCoro(self.__status_proc)
        self.__client_coro.register('discoro_scheduler')
        self.asyncoro.discover_peers(port=self._node_port)

    def status(self):
        pending = sum(node.ncoros for node in self._nodes.values())
        servers = functools.reduce(operator.add, [list(node.servers.keys())
                                                  for node in self._nodes.values()], [])
        return {'Client': self._cur_computation._pulse_coro.location if self._cur_computation else '',
                'Pending': pending, 'Nodes': list(self._nodes.keys()), 'Servers': servers
                }

    def print_status(self):
        status = self.status()
        print('')
        print('  Client: %s' % status['Client'])
        print('  Pending: %s' % status['Pending'])
        print('  nodes: %s' % len(status['Nodes']))
        print('  servers: %s' % len(status['Servers']))

    def __status_proc(self, coro=None):
        coro.set_daemon()
        coro.register('discoro_status')
        self.asyncoro.peer_status(coro)
        while 1:
            msg = yield coro.receive()
            if isinstance(msg, asyncoro.MonitorException):
                rcoro = msg.args[0]
                if not isinstance(rcoro, Coro):
                    logger.warning('ignoring invalid rcoro %s', type(rcoro))
                    continue
                node = self._nodes.get(rcoro.location.addr, None)
                if not node:
                    logger.warning('node %s is invalid', rcoro.location.addr)
                    continue
                server = node.servers.get(rcoro.location, None)
                if not server:
                    logger.warning('server "%s" is invalid', rcoro.location)
                    continue
                if server.rcoros.pop(rcoro, None) is None:
                    # Due to 'yield' used to create rcoro, scheduler may not
                    # have updated self._rcoros before the coroutine's
                    # MonitorException is received, so put it in
                    # 'askew_results'. The scheduling coroutine will resend it
                    # when it receives rcoro
                    server.askew_results[rcoro] = msg
                    continue

                if self._cur_computation and self._cur_computation.status_coro:
                    if len(msg.args) > 2:
                        msg.args = (msg.args[0], msg.args[1])
                    self._cur_computation.status_coro.send(msg)
                node.ncoros -= 1
                node.load = float(node.ncoros) / len(node.servers)

            elif isinstance(msg, asyncoro.PeerStatus):
                if msg.status == asyncoro.PeerStatus.Online:
                    if msg.name.endswith('_proc-0'):
                        SysCoro(self.__discover_node, msg)
                else:
                    # msg.status == asyncoro.PeerStatus.Offline
                    node = self._nodes.get(msg.location.addr, None)
                    if node:
                        server = node.servers.pop(msg.location, None)
                        if server:
                            SysCoro(self.__close_server, server, self._cur_computation)
                        elif node.coro and node.coro.location == msg.location:
                            # TODO: inform scheduler / client
                            self._nodes.pop(msg.location.addr)
                        # if not node.servers:
                        #     self._nodes.pop(msg.location.addr)
                    elif (self._cur_computation and
                          self._cur_computation._pulse_coro.location == msg.location):
                        logger.warning('Client %s terminated; closing computation %s',
                                       msg.location, self.__cur_client_auth)
                        SysCoro(self.__close_computation)

            elif isinstance(msg, dict):  # message from a node's server
                status = msg.get('status', None)
                now = time.time()
                if status == 'pulse':
                    location = msg.get('location', None)
                    if not isinstance(location, asyncoro.Location):
                        continue
                    node = self._nodes.get(location.addr, None)
                    if node:
                        for server in node.servers.values():
                            server.last_pulse = now
                        node_status = msg.get('node_status', None)
                        if (node_status and self._cur_computation and
                           self._cur_computation.status_coro):
                            self._cur_computation.status_coro.send(node_status)

                elif status == Scheduler.ServerInitialized:
                    rcoro = msg.get('coro', None)
                    if not isinstance(rcoro, asyncoro.Coro):
                        continue
                    if (self._cur_computation and
                        self._cur_computation._auth == msg.get('auth', None)):

                        node = self._nodes.get(rcoro.location.addr, None)
                        if node and (node.status == Scheduler.NodeInitialized or
                                     node.status == Scheduler.NodeDiscovered):
                            server = node.servers.get(rcoro.location, None)
                            if not server:
                                server = Scheduler._Server(msg.get('name', None), rcoro.location)
                                server.coro = rcoro
                                node.servers[rcoro.location] = server
                            server.status = Scheduler.ServerInitialized
                            server.last_pulse = now
                            if node.status != Scheduler.NodeInitialized:
                                node.status = Scheduler.NodeInitialized
                                if self._cur_computation.status_coro:
                                    info = DiscoroNodeInfo(node.name, node.addr, node.cpus,
                                                           node.platform, node.avail_info)
                                    self._cur_computation.status_coro.send(
                                        DiscoroStatus(node.status, info))
                            if self._cur_computation.status_coro:
                                self._cur_computation.status_coro.send(DiscoroStatus(server.status,
                                                                                     server.location))
                            if self._cur_computation._peers_communicate:
                                server.coro.send({'req': 'peers',
                                                  'auth': self._cur_computation._auth,
                                                  'peers': list(self.__server_locations)})
                                self.__server_locations.add(server.location)

                elif status in (Scheduler.ServerClosed, Scheduler.ServerDisconnected):
                    location = msg.get('location', None)
                    if not isinstance(location, asyncoro.Location):
                        continue
                    node = self._nodes.get(location.addr, None)
                    if not node:
                        continue
                    if status == Scheduler.ServerClosed:
                        server = node.servers.get(location, None)
                        if server:
                            server.status = Scheduler.ServerIgnore
                    else:
                        server = node.servers.pop(location, None)
                        if not node.servers:
                            node.status = Scheduler.NodeIgnore
                    if not server:
                        continue
                    if self._cur_computation:
                        if self._cur_computation._peers_communicate:
                            self.__server_locations.discard(server.location)
                            # TODO: inform other servers
                        if self._cur_computation.status_coro:
                            self._cur_computation.status_coro.send(DiscoroStatus(status, location))
                            if not node.servers:
                                self._cur_computation.status_coro.send(
                                    DiscoroStatus(Scheduler.NodeClosed, node.addr))
                else:
                    logger.warning('Ignoring invalid status message: %s', status)
            else:
                logger.warning('invalid status message ignored')

    def __node_allocate(self, node):
        for node_allocate in self.__cur_node_allocations:
            cpus = node_allocate.allocate(node.addr, node.name, node.platform, node.cpus,
                                          node.avail_info.memory, node.avail_info.disk)
            if cpus < 0:
                continue
            return min(cpus, node.cpus)
        return node.cpus

    def __get_node_info(self, node, coro=None):
        if not node.coro:
            logger.warning('Node %s is not valid', node.addr)
            raise StopIteration
        node.coro.send({'req': 'discoro_node_info', 'client': coro})
        node_info = yield coro.receive(timeout=MsgTimeout)
        if not node_info:
            self._nodes.pop(node.addr, None)
            asyncoro.Coro(asyncoro.AsynCoro.instance().close_peer, node.coro.location)
            raise StopIteration
        node.name = node_info.name
        node.cpus = node_info.cpus
        node.platform = node_info.platform.lower()
        node.avail_info = node_info.avail_info
        if not self._cur_computation:
            raise StopIteration
        cpus = self.__node_allocate(node)
        if not cpus:
            node.status = Scheduler.NodeIgnore
            raise StopIteration
        if self._cur_computation:
            SysCoro(self.__reserve_node, node, cpus)
        else:
            node.status = Scheduler.NodeDiscovered

    def __reserve_node(self, node, cpus, coro=None):
        computation = self._cur_computation
        if not computation:
            raise StopIteration(-1)
        node.coro.send({'req': 'computation', 'computation': computation, 'cpus': cpus,
                        'status_coro': self.__status_coro, 'client': coro})
        cpus = yield coro.receive(timeout=computation.timeout, alarm_value=-1)
        if not cpus:
            logger.warning('setup of %s failed', node.coro)
            # self._nodes.pop(node.addr, None)
            SysCoro(self.__close_node, node, computation)
            raise StopIteration(-1)
        node.cpus = cpus
        if not self._cur_computation:
            raise StopIteration(0)

        xfer_files = computation._xfer_files
        timeout = computation.timeout
        for xf, dst, sep in xfer_files:
            reply = yield self.asyncoro.send_file(node.coro.location, xf, dir=dst,
                                                  timeout=timeout)
            if reply < 0:
                logger.debug('failed to transfer file %s: %s', xf, reply)
                SysCoro(self.__close_node, node, computation)
                raise StopIteration(-1)

        if self.__cur_node_available:
            if self._shared:
                node.status = Scheduler.NodeDiscovered
                info = DiscoroNodeInfo(node.name, node.addr, node.cpus,
                                       node.platform, node.avail_info)
                computation.status_coro.send(DiscoroStatus(node.status, info))
                raise StopIteration(0)
            try:
                params = yield asyncoro.Coro(self.__cur_node_available, node.avail_info).finish()
                assert params is not None
            except:
                node.status = Scheduler.NodeIgnore
                SysCoro(self.__close_node, node, computation)
                raise StopIteration(-1)
            else:
                if not isinstance(params, tuple):
                    if hasattr(params, '__iter__'):
                        params = tuple(params)
                    else:
                        params = (params,)
        else:
            params = ()

        if computation._node_setup:
            if (yield node.coro.deliver({'req': 'setup', 'params': params,
                                         'auth': computation._auth})) != 1:
                SysCoro(self.__close_node, node, computation)
                raise StopIteration(-1)
        node.status = Scheduler.NodeInitialized
        if computation.status_coro:
            info = DiscoroNodeInfo(node.name, node.addr, node.cpus, node.platform,
                                   node.avail_info)
            computation.status_coro.send(DiscoroStatus(node.status, info))

    def __discover_node(self, msg, coro=None):
        for _ in range(10):
            rcoro = yield Coro.locate('discoro_node', location=msg.location,
                                      timeout=MsgTimeout)
            if not isinstance(rcoro, Coro):
                yield coro.sleep(0.1)
                continue
            node = self._nodes.get(msg.location.addr, None)
            if node and node.coro == rcoro:
                raise StopIteration

            if not node:
                node = Scheduler._Node(msg.name, msg.location.addr)
                self._nodes[msg.location.addr] = node
            node.coro = rcoro
            SysCoro(self.__get_node_info, node)
            raise StopIteration

    def __timer_proc(self, coro=None):
        coro.set_daemon()
        server_check = client_pulse = last_ping = time.time()
        async_scheduler = coro.scheduler()
        while 1:
            try:
                msg = yield coro.receive(timeout=self.__pulse_interval)
            except GeneratorExit:
                break
            now = time.time()
            if (now - client_pulse) > self.__pulse_interval and self.__cur_client_auth:
                if self._cur_computation._pulse_coro.send('pulse') == 0:
                    client_pulse = now
                elif (self._cur_computation.zombie_period and
                      (now - client_pulse) > self._cur_computation.zombie_period):
                    logger.warning('Closing zombie computation %s', self.__cur_client_auth)
                    SysCoro(self.__close_computation)

            if (self._cur_computation and self._cur_computation.zombie_period and
               (now - server_check) > self._cur_computation.zombie_period):
                server_check = now
                for node in self._nodes.values():
                    if node.status != Scheduler.NodeInitialized:
                        continue
                    for server in node.servers.values():
                        if server.status != Scheduler.ServerInitialized:
                            continue
                        if (now - server.last_pulse) > self._cur_computation.zombie_period:
                            logger.warning('discoro server %s is zombie!', server.location)
                            SysCoro(self.__close_server, server, self._cur_computation)

            if self.__ping_interval and ((now - last_ping) > self.__ping_interval):
                last_ping = now
                SysCoro(async_scheduler.discover_peers, port=self._node_port)

    def __run(self, func, client):
        host = None
        load = None
        for node in self._nodes.values():
            if node.status != Scheduler.NodeInitialized:
                continue
            if load is None or node.load < load:
                host = node
                load = node.load
        if host:
            yield host.run(func, self._cur_computation, client)
        else:
            client.send(None)

    @staticmethod
    def auth_code():
        # TODO: use uuid?
        return hashlib.sha1(''.join(hex(_)[2:] for _ in os.urandom(10)).encode()).hexdigest()

    def __scheduler_proc(self, nodes, coro=None):
        coro.set_daemon()
        for node in nodes:
            yield asyncoro.AsynCoro.instance().peer(node, broadcast=True)
        while not self.__terminate:
            if self._cur_computation:
                self.__sched_event.clear()
                yield self.__sched_event.wait()
                continue

            self._cur_computation, client = yield coro.receive()

            self.__pulse_interval = self._cur_computation._pulse_interval
            self.__ping_interval = self._cur_computation._ping_interval

            self.__cur_client_auth = self._cur_computation._auth
            self._cur_computation._auth = Scheduler.auth_code()
            self.__cur_node_allocations = self._cur_computation._node_allocations
            self._cur_computation._node_allocations = []
            self.__cur_node_available = self._cur_computation._node_available
            self._cur_computation._node_available = None

            logger.debug('Computation %s / %s scheduled', self.__cur_client_auth,
                         self._cur_computation._auth)
            msg = {'resp': 'scheduled', 'auth': self.__cur_client_auth}
            if (yield client.deliver(msg, timeout=min(self._cur_computation.timeout, MsgTimeout))) != 1:
                logger.warning('client not reachable?')
                self._cur_client_auth = None
                self._cur_computation = None
                continue

            if self._cur_computation.status_coro:
                status_msg = DiscoroStatus(Scheduler.ComputationScheduled, id(self._cur_computation))
                self._cur_computation.status_coro.send(status_msg)

            self.__timer_coro.send(None)
            if self._cur_computation:
                for node in self._nodes.values():
                    SysCoro(self.__get_node_info, node)
        self.__scheduler_coro = None

    def __client_proc(self, coro=None):
        coro.set_daemon()
        computations = {}
        while not self.__terminate:
            msg = yield coro.receive()
            if not isinstance(msg, dict):
                continue
            req = msg.get('req', None)
            client = msg.get('client', None)
            auth = msg.get('auth', None)
            if not isinstance(client, Coro):
                logger.warning('Ignoring invalid client request "%s"', req)
                continue

            if req == 'run':
                func = msg.get('func', None)
                if not func or self.__cur_client_auth != auth:
                    logger.warning('Ignoring invalid request to run computation')
                    client.send(None)
                    continue
                where = msg.get('where', None)
                if not where:
                    SysCoro(self.__run, func, client)
                elif isinstance(where, str):
                    node = self._nodes.get(where, None)
                    if node:
                        SysCoro(node.run, func, self._cur_computation, client)
                    else:
                        client.send(None)
                elif isinstance(where, asyncoro.Location):
                    node = self._nodes.get(where.addr)
                    if node:
                        server = node.servers.get(where)
                        if server:
                            SysCoro(server.run, func, self._cur_computation, node, client)
                        else:
                            client.send(None)
                    else:
                        client.send(None)
                else:
                    client.send(None)

            elif req == 'run_each':
                where = msg.get('where', None)
                func = msg.get('func', None)
                if not func or self.__cur_client_auth != auth:
                    logger.warning('Ignoring invalid request to run computation')
                    where = None
                if where == 'node':
                    nodes = [node for node in self._nodes.values()
                             if node.status == Scheduler.NodeInitialized]
                    if (yield client.deliver(len(nodes), self._cur_computation.timeout)) != 1:
                        continue
                    for node in nodes:
                        SysCoro(node.run, func, self._cur_computation, client)
                elif where == 'server':
                    node_servers = [(node, server) for node in self._nodes.values()
                                    if node.status == Scheduler.NodeInitialized
                                    for server in node.servers.values()
                                    if server.status == Scheduler.ServerInitialized]
                    if (yield client.deliver(len(node_servers), self._cur_computation.timeout)) != 1:
                        continue
                    for node, server in node_servers:
                        SysCoro(server.run, func, self._cur_computation, node, client)
                else:
                    node = self._nodes.get(where)
                    if node and node.status == Scheduler.NodeInitialized:
                        servers = [server for server in node.servers.values()
                                   if server.status == Scheduler.ServerInitialized]
                    else:
                        servers = []
                    if (yield client.deliver(len(servers), self._cur_computation.timeout)) != 1:
                        continue
                    for server in servers:
                        SysCoro(server.run, func, self._cur_computation, node, client)

            elif req == 'schedule':
                try:
                    computation = asyncoro.deserialize(msg['computation'])
                    assert isinstance(computation, Computation) or \
                        computation.__class__.__name__ == 'Computation'
                    assert isinstance(computation._pulse_coro, Coro)
                    if computation._pulse_coro.location == self.asyncoro.location:
                        computation._pulse_coro._id = int(computation._pulse_coro._id)
                        if computation.status_coro:
                            computation.status_coro._id = int(computation.status_coro._id)
                    assert isinstance(computation._pulse_interval, (float, int))
                    assert (MinPulseInterval <= computation._pulse_interval <= MaxPulseInterval)
                    if computation._ping_interval:
                        assert computation._ping_interval >= MinPulseInterval
                except:
                    logger.warning('ignoring invalid computation request')
                    client.send(None)
                    continue
                while 1:
                    computation._auth = Scheduler.auth_code()
                    if not os.path.exists(os.path.join(self.__dest_path, computation._auth)):
                        break
                try:
                    os.mkdir(os.path.join(self.__dest_path, computation._auth))
                except:
                    logger.debug('Could not create "%s"',
                                 os.path.join(self.__dest_path, computation._auth))
                    client.send(None)
                    continue
                # TODO: save it on disk instead
                computations[computation._auth] = computation
                client.send(computation._auth)

            elif req == 'await':
                computation = computations.pop(auth, None)
                if not computation:
                    client.send(None)
                    continue
                if computation._pulse_coro.location.addr != self.asyncoro.location.addr:
                    computation._xfer_files = [(os.path.join(self.__dest_path, computation._auth,
                                                             os.path.join(*(dst.split(sep))),
                                                             xf.split(sep)[-1]),
                                                os.path.join(*(dst.split(sep))), os.sep)
                                               for xf, dst, sep in computation._xfer_files]
                for xf, dst, sep in computation._xfer_files:
                    if not os.path.isfile(xf):
                        logger.warning('File "%s" for computation %s is not valid',
                                       xf, computation._auth)
                        computation = None
                        break
                if computation is None:
                    client.send(None)
                else:
                    # TODO: allow zombie_period to be set?
                    if self.__zombie_period:
                        computation.zombie_period = self.__zombie_period
                    self.__scheduler_coro.send((computation, client))
                    self.__sched_event.set()

            elif req == 'close_computation':
                if self.__cur_client_auth == auth:
                    SysCoro(self.__close_computation, client=client)
                else:
                    computation = computations.pop(auth, None)
                    if computation:
                        computation_path = os.path.join(self.__dest_path, auth)
                        if os.path.isdir(computation_path):
                            shutil.rmtree(computation_path, ignore_errors=True)
                    else:
                        logger.warning('Ignoring invalid request to close computation')
                    client.send('closed')

            elif req == 'nodes_list':
                # TODO: allowed to query anytime, even if current
                # computation is not the one querying?
                if self.__cur_client_auth == auth:
                    nodes = [node.addr for node in self._nodes.values()
                             if node.status == Scheduler.NodeInitialized]
                else:
                    nodes = []
                client.send(nodes)

            elif req == 'servers_list':
                # TODO: allowed to query anytime, even if current
                # computation is not the one querying?
                if self.__cur_client_auth == auth:
                    servers = [server.location for node in self._nodes.values()
                               if node.status == Scheduler.NodeInitialized
                               for server in node.servers.values()
                               if server.status == Scheduler.ServerInitialized]
                else:
                    servers = []
                client.send(servers)

            elif req == 'setup_node':
                addr = msg.get('addr', None)
                params = msg.get('params', None)
                if isinstance(addr, asyncoro.Location):
                    addr = addr.addr
                node = self._nodes.get(addr, None)
                status = -1
                if (self.__cur_client_auth == auth and node and
                    node.status == Scheduler.NodeDiscovered and self._cur_computation._node_setup):
                    if (yield node.coro.deliver({'req': 'setup', 'params': params,
                                                 'auth': self._cur_computation._auth})) == 1:
                        status = 0
                    else:
                        SysCoro(self.__close_node, node, self._cur_computation)
                # client.send(status)

            else:
                logger.warning('Ignoring invalid client request "%s"', req)

    def __close_node(self, node, computation, coro=None):
        if not computation or node.status != Scheduler.NodeInitialized:
            logger.warning('Closing node %s ignored: %s', node.addr, node.status)
            raise StopIteration(-1)
        close_coros = [SysCoro(self.__close_server, server, computation)
                       for server in node.servers.values()]
        for close_coro in close_coros:
            yield close_coro.finish()
        if node.coro:
            node.coro.send({'req': 'release', 'auth': computation._auth, 'client': coro})
            yield coro.receive(timeout=computation.timeout)

    def __close_server(self, server, computation, coro=None):
        if self.__server_locations:
            self.__server_locations.discard(server.location)
            # TODO: inform other servers
        if server.status != Scheduler.ServerInitialized or not computation:
            logger.debug('Closing server %s ignored: %s', server.location, server.status)
            raise StopIteration(-1)
        node = self._nodes.get(server.location.addr, None)
        if not node:
            raise StopIteration(-1)
        disconnected = server.location not in node.servers
        if disconnected:
            if computation and computation.status_coro:
                computation.status_coro.send(DiscoroStatus(Scheduler.ServerDisconnected,
                                                           server.location))
        else:
            server.coro.send({'req': 'close', 'auth': computation._auth, 'client': coro})
            yield coro.receive(timeout=computation.timeout if computation else MsgTimeout)
        if server.rcoros:  # wait a bit for server to terminate coros
            for _ in range(5):
                yield coro.sleep(0.1)
                if not server.rcoros:
                    break
        if server.rcoros:
            logger.warning('%s coros running at %s', len(server.rcoros), server.location)
            if computation and computation.status_coro:
                for rcoro in server.rcoros.values():
                    status = asyncoro.MonitorException(rcoro, (Scheduler.ServerClosed, None))
                    computation.status_coro.send(status)

        server.status = Scheduler.ServerClosed
        server.xfer_files = []
        server.rcoros.clear()
        server.askew_results.clear()
        node.servers.pop(server.location, None)
        server.coro = None

        if computation and computation.status_coro:
            computation.status_coro.send(DiscoroStatus(server.status, server.location))
        if disconnected and not node.servers:
            node.ncoros = 0
            node.load = 0.0
            node.status = Scheduler.NodeClosed
            if computation and computation.status_coro:
                computation.status_coro.send(DiscoroStatus(node.status, node.addr))
        raise StopIteration(0)

    def __close_computation(self, client=None, coro=None):
        self.__server_locations.clear()
        if self.__cur_client_auth:
            computation_path = os.path.join(self.__dest_path, self.__cur_client_auth)
            if os.path.isdir(computation_path):
                shutil.rmtree(computation_path, ignore_errors=True)
        if self._cur_computation:
            close_coros = [SysCoro(self.__close_node, node, self._cur_computation)
                           for node in self._nodes.values()]
            for close_coro in close_coros:
                yield close_coro.finish()
        if self._cur_computation and self._cur_computation.status_coro:
            self._cur_computation.status_coro.send(DiscoroStatus(Scheduler.ComputationClosed,
                                                                 id(self._cur_computation)))
        self.__cur_client_auth = self._cur_computation = None
        self.__sched_event.set()
        if client:
            client.send('closed')
        raise StopIteration(0)

    def close(self, coro=None):
        """Close current computation and quit scheduler.

        Must be called with 'yield' as 'yield scheduler.close()' or as
        coroutine.
        """
        if self.__terminate:
            raise StopIteration(-1)
        self.__terminate = True
        self.__sched_event.set()
        yield self.__close_computation(coro=coro)
        raise StopIteration(0)


if __name__ == '__main__':
    """The scheduler can be started either within a client program (if no other
    client programs use the nodes simultaneously), or can be run on a node with
    the options described below (usually no options are necessary, so the
    scheduler can be strated with just 'discoro.py')
    """

    import logging
    import argparse
    import signal
    try:
        import readline
    except ImportError:
        pass

    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--ip_addr', dest='node', default=None,
                        help='IP address or host name of this node')
    parser.add_argument('--ext_ip_addr', dest='ext_ip_addr', default=None,
                        help='External IP address to use (needed in case of NAT firewall/gateway)')
    parser.add_argument('-u', '--udp_port', dest='udp_port', type=int, default=51350,
                        help='UDP port number to use')
    parser.add_argument('-t', '--tcp_port', dest='tcp_port', type=int, default=0,
                        help='TCP port number to use')
    parser.add_argument('-n', '--name', dest='name', default=None,
                        help='(symbolic) name given to schduler')
    parser.add_argument('--dest_path', dest='dest_path', default=None,
                        help='path prefix to where files sent by peers are stored')
    parser.add_argument('--max_file_size', dest='max_file_size', default=None, type=int,
                        help='maximum file size of any file transferred')
    parser.add_argument('-s', '--secret', dest='secret', default='',
                        help='authentication secret for handshake with peers')
    parser.add_argument('--certfile', dest='certfile', default=None,
                        help='file containing SSL certificate')
    parser.add_argument('--keyfile', dest='keyfile', default=None,
                        help='file containing SSL key')
    parser.add_argument('--node', action='append', dest='nodes', default=[],
                        help='additional remote nodes (names or IP address) to use')
    parser.add_argument('--zombie_period', dest='zombie_period', type=int,
                        default=(10 * MaxPulseInterval),
                        help='maximum time in seconds computation is idle')
    parser.add_argument('-d', '--debug', action='store_true', dest='loglevel', default=False,
                        help='if given, debug messages are printed')
    parser.add_argument('--clean', action='store_true', dest='clean', default=False,
                        help='if given, files copied from or generated by clients will be removed')
    parser.add_argument('--discoronode_port', dest='discoronode_port', type=int, default=51351,
                        help='UDP port number used by discoronode')
    parser.add_argument('--daemon', action='store_true', dest='daemon', default=False,
                        help='if given, input is not read from terminal')
    config = vars(parser.parse_args(sys.argv[1:]))
    del parser

    if config['zombie_period'] and config['zombie_period'] < MaxPulseInterval:
        raise Exception('zombie_period must be >= %s' % MaxPulseInterval)

    if not config['name']:
        config['name'] = 'discoro_scheduler'

    if config['loglevel']:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    del config['loglevel']

    daemon = config.pop('daemon', False)
    _discoro_scheduler = Scheduler(**config)
    _discoro_scheduler._shared = True
    del config

    def sighandler(signum, frame):
        Coro(_discoro_scheduler.close).value()
        raise KeyboardInterrupt

    try:
        signal.signal(signal.SIGHUP, sighandler)
        signal.signal(signal.SIGQUIT, sighandler)
    except:
        pass
    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGABRT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)
    del sighandler

    if not daemon:
        try:
            if os.getpgrp() != os.tcgetpgrp(sys.stdin.fileno()):
                daemon = True
        except:
            pass

    if daemon:
        del daemon
        while 1:
            try:
                time.sleep(3600)
            except:
                break
    else:
        del daemon
        while 1:
            try:
                _discoro_cmd = input(
                    '\n\nEnter "quit" or "exit" to terminate discoro scheduler\n'
                    '      "status" to show status of scheduler: '
                    )
            except:
                break
            _discoro_cmd = _discoro_cmd.strip().lower()
            if _discoro_cmd in ('quit', 'exit'):
                Coro(_discoro_scheduler.close).value()
                break
            if _discoro_cmd == 'status':
                _discoro_scheduler.print_status()

    logger.info('terminating discoro scheduler')



import sys
import os
import glob
from setuptools import setup

if sys.version_info.major == 3:
    base_dir = 'py3'
else:
    assert sys.version_info.major == 2
    assert sys.version_info.minor >= 7
    base_dir = 'py2'

setup(
    name='asyncoro',
    version='4.3.3',
    description='Python framework for concurrent, distributed, asynchronous network programming with coroutines, asynchronous completions and message passing.',
    long_description=open('README.rst').read(),
    keywords='concurrent, distributed, asynchronous network programming, coroutines, message passing',
    url='http://asyncoro.sourceforge.net',
    author='Giridhar Pemmasani',
    author_email='pgiri@yahoo.com',
    package_dir={'':base_dir},
    packages=['asyncoro'],
    package_data = {
        'asyncoro' : ['data/*', 'doc/*', 'examples/*'],
    },
    scripts=[os.path.join(base_dir, 'asyncoro', script)
             for script in ['discoro.py', 'discoronode.py']],
    license='MIT',
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        ]
    )



#!/usr/bin/env python

# to be used with 'rci_monitor_server.py'
# client requests execution of coroutines on (remote) server.

import sys, random
# import disasyncoro to use distributed version of AsynCoro
import asyncoro.disasyncoro as asyncoro

def monitor_proc(n, coro=None):
    # this coro gets exceptions from (remote) coroutines created in coro1
    done = 0
    while True:
        msg = yield coro.receive()
        if isinstance(msg, asyncoro.MonitorException):
            rcoro = msg.args[0]
            value_type, value = msg.args[1]
            if value_type == StopIteration:
                asyncoro.logger.debug('RCI %s finished with: %s', rcoro, value)
            else:
                asyncoro.logger.debug('RCI %s failed with: %s', rcoro, value.args[0])
            done += 1
            if done == n:
                break
        else:
            asyncoro.logger.warning('ignoring invalid message')

def rci_test(coro=None):
    # if server is on remote network, automatic discovery won't work,
    # so add it explicitly
    # yield scheduler.peer('192.168.21.5')

    # find where 'rci_1' is registered (with given name in any
    # known peer)
    rci1 = yield asyncoro.RCI.locate('rci_1')
    print('RCI is at %s' % rci1.location)
    # alternately, location can be explicitly created with
    # asyncoro.Location or obtained with 'locate' of asyncoro etc.
    loc = yield scheduler.locate('server')
    rci1 = yield asyncoro.RCI.locate('rci_1', loc)
    print('RCI is at %s' % rci1.location)

    n = 5
    monitor = asyncoro.Coro(monitor_proc, n)
    for x in range(n):
        rcoro = yield rci1('test%s' % x, b=x)
        asyncoro.logger.debug('RCI %s created' % rcoro)
        # set 'monitor' as monitor for this coroutine
        yield monitor.monitor(rcoro)
        # send a message
        rcoro.send('msg:%s' % x)
        yield coro.sleep(random.uniform(0, 1))

asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
scheduler = asyncoro.AsynCoro(name='client', secret='test')
asyncoro.Coro(rci_test)



# Run 'discoronode.py' program to start processes to execute
# computations sent by this client, along with this program.

# This program is similar to 'discoro_client1.py', but instead of using
# RemoteCoroScheduler, it uses status message notifications from discoro
# scheduler to submit jobs at remote discoronode processes.

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *

# this generator function is sent to remote server to run coroutines there
def rcoro_proc(n, coro=None):
    yield coro.sleep(n)
    raise StopIteration(n)


# Instead of using RemoteCoroScheduler (which is easier), in this example status
# messages from discoro scheduler are used to start remote coroutines and get
# their exit status
def status_proc(computation, njobs, coro=None):
    npending = njobs

    # in this example at most one coroutine is submitted at a server; depending
    # on computation / needs, many coroutines can be simlutaneously submitted /
    # running at a server.
    while True:
        msg = yield coro.receive()
        if isinstance(msg, DiscoroStatus):
            # print('Status: %s / %s' % (msg.info, msg.status))
            if msg.status == Scheduler.ServerInitialized and njobs > 0: # submit a job
                n = random.uniform(5, 10)
                rcoro = yield computation.run_at(msg.info, rcoro_proc, n)
                if isinstance(rcoro, asyncoro.Coro):
                    print('  rcoro_proc started at %s with %s' % (rcoro.location, n))
                    njobs -= 1
        elif isinstance(msg, asyncoro.MonitorException):
            # previously submitted remote coroutine finished
            rcoro = msg.args[0]
            if msg.args[1][0] == StopIteration: # exit status type
                print('      rcoro_proc at %s finished with %s' % (rcoro.location, msg.args[1][1]))
            else:
                print('      rcoro_proc at %s failed: %s / %s' %
                      (rcoro.location, msg.args[1][0], msg.args[1][1]))

            npending -= 1
            if npending == 0:
                break
            if njobs > 0: # submit another job
                n = random.uniform(5, 10)
                rcoro = yield computation.run_at(rcoro.location, rcoro_proc, n)
                if isinstance(rcoro, asyncoro.Coro):
                    print('  rcoro_proc started at %s with %s' % (rcoro.location, n))
                    njobs -= 1

def client_proc(computation, njobs, coro=None):
    computation.status_coro = asyncoro.Coro(status_proc, computation, njobs)
    # since RemoteCoroScheduler is not used, schedule computation
    if (yield computation.schedule()):
        raise Exception('Failed to schedule computation')
    # wait for jobs to be created and finished
    yield computation.status_coro.finish()
    yield computation.close()


if __name__ == '__main__':
    import random
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # if scheduler is not already running (on a node as a program),
    # start it (private scheduler):
    Scheduler()
    computation = Computation([rcoro_proc])
    # run 10 jobs
    asyncoro.Coro(client_proc, computation, 10)



# This example uses status messages and message passing to run 'setup' coroutine
# at remote process to prepare it for processing jobs.

import asyncoro.disasyncoro as asyncoro
from asyncoro.discoro import *
from asyncoro.discoro_schedulers import RemoteCoroScheduler

# this generator function is sent to remote discoro servers to run
# coroutines there
def compute(i, n, coro=None):
    import time
    yield coro.sleep(n)
    raise StopIteration((i, coro.location, time.asctime())) # result of 'compute' is current time

# client (local) coroutine submits computations
def client_proc(computation, njobs, coro=None):
    # execute n jobs (coroutines) and get their results. Number of jobs created
    # can be more than number of server processes available; the scheduler will
    # use as many processes as necessary/available, running one job at a server
    # process

    # arguments must correspond to arguments for computaiton; multiple arguments
    # (as in this case) can be given as tuples
    args = [(i, random.uniform(2, 5)) for i in range(njobs)]
    results = yield rcoro_scheduler.map_results(compute, args)
    # Coroutines may not be executed in the order of given list of args, but
    # results would be in the same order of given list of args
    for result in results:
        print('    result for %d from %s: %s' % result)
    # yield rcoro_scheduler.finish(close=True)


if __name__ == '__main__':
    import random, sys
    # asyncoro.logger.setLevel(asyncoro.Logger.DEBUG)
    # if scheduler is not already running (on a node as a program), start
    # private scheduler:
    Scheduler()
    # package computation fragments
    computation = Computation([compute])
    # use RemoteCoroScheduler to start coroutines at servers (should be done
    # before scheduling computation)
    rcoro_scheduler = RemoteCoroScheduler(computation)
    # run 10 (or given number of) jobs
    asyncoro.Coro(client_proc, computation, 10 if len(sys.argv) < 2 else int(sys.argv[1]))
